/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef FILTER_FAST_INTERSECTOR_HH_
# define FILTER_FAST_INTERSECTOR_HH_

# include <utility>

# include <wipal/tool/hash.hh>
# include <wipal/tool/valued_iterator.hh>
# include <wipal/tool/microseconds.hh>
# include <wipal/tool/iterable.hh>
# include <wipal/wifi/addr_mapping.hh>

# include "fast_intersector_fwd.hh"

namespace wpl
{

  namespace tool
  {

    template <class I1, class I2, class B1, class B2>
    struct types< wpl::filter::internals::
      fast_intersector_iterator<I1, I2, B1, B2> >
    {
      typedef
      std::pair<typename I1::value_type, typename I2::value_type>
      value_type;

      typedef
      wpl::filter::fast_intersector<I1, I2, B1>
      iterable_type;
    };

    template <class I1, class I2, class B>
    struct types< wpl::filter::fast_intersector<I1, I2, B> >
    {
      typedef wpl::filter::internals::
      fast_intersector_iterator<I1, I2, B, bottom>
      iterator;
    };

  } // End of namespace wpl::tool.

  namespace filter
  {

    namespace internals
    {
      template <class I1, class I2, class B, class Bottom = tool::bottom>
      struct fast_intersector_iterator:
	WP_INHERIT(public tool::valued_iterator,
		   fast_intersector_iterator<I1, I2, B, Bottom>)
      {

	/// @name Static hierarchies related types
	//@{
	typedef WP_GET_EXACT(Bottom,
			     fast_intersector_iterator<I1, I2, B, Bottom>)
	  exact_type;
	typedef tool::valued_iterator<exact_type>	super_type;
	typedef WP_TYPE(value_type, exact_type)		value_type;
	typedef WP_TYPE(iterable_type, exact_type)	iterable_type;
	typedef typename iterable_type::arrivals_type	arrivals_type;
	typedef typename arrivals_type::const_iterator	arrivals_iterator;
	//@}

	/// tool::iterable compatible constructor.
	fast_intersector_iterator(const iterable_type& i, bool end);

	/// @name Abstract method implementations
	//@{
	bool		equal(const fast_intersector_iterator& rhs) const;
	void		increment();
	//@}

	/// Get the number of processed frames inside trace 2.
	unsigned	rhs_frame_count() const;

	/// Get the number of processed unique frames inside trace 2.
	unsigned	rhs_uframe_count() const;

      private:
	enum
	{
	  /**
	   * Generate a warning when the interval between two
	   * reference frames exceed @c dt_warn_threshold
	   * microseconds.
	   *
	   * Arbitrarily set to 15 minutes.
	   */
	  dt_warn_threshold = 15 * 60 * 1000 * 1000,

	  /**
	   * Ignore reference frames when the drift between them
	   * exceed @c drift_ignore_threshold microseconds.
	   *
	   * Set to 20 milliseconds.
	   *
	   * Rationale: This is related to some traces with timer
	   *            anomalies we experienced.  Merging these
	   *            traces duplicates some reference frames in
	   *            the output.  In that case, output duplicates
	   *            are spaced by 32767 microseconds.  That in
	   *            turn generates frames with a similar drift
	   *            in subsequent merges.  Thus we need a lower
	   *            threshold.  We arbitrarily pick 20000
	   *            microseconds.
	   */
	  drift_warn_threshold = 20 * 1000,
	};


	bool	check();
	bool	check_drift(const tool::microseconds& dt1,
			    const tool::microseconds& dt2,
			    const typename I1::value_type& frame1,
			    const typename I2::value_type& frame2);
	void	check_consistency(const tool::microseconds& dt1,
				  const tool::microseconds& dt2,
				  const typename I1::value_type& frame1,
				  const typename I2::value_type& frame2) const;

	void	ignore(const typename I1::value_type& frame1,
		       const typename I2::value_type& frame2);
	void	dont_ignore(const typename I1::value_type& frame1,
			    const typename I2::value_type& frame2);

	/// Maintain statistics about ignored frames.
	struct ignore_info
	{
	  ignore_info(const tool::microseconds&	stamp_rhs,
		      const tool::microseconds&	stamp_lhs,
		      const unsigned		total_rhs,
		      const unsigned		total_lhs);

	  /// Timestamp of the first ignored frame in trace 1.
	  tool::microseconds	stamp1_;

	  /// Timestamp of the first ignored frame in trace 2.
	  tool::microseconds	stamp2_;

	  /// Index of the first ignored frame in trace 1.
	  unsigned		total1_;

	  /// Index of the first ignored frame in trace 2.
	  unsigned		total2_;
	};

	const iterable_type*		iterable_;
	I2				next_;
	boost::optional<ignore_info>	ign_;

	// We use a lookahead of one frame so we can filter the
	// penultimate reference out if it turns out to be dubious.

	typename super_type::optional_value	lookahead_;

# ifdef WP_ENABLE_INFO
	unsigned		rhs_total_;
	unsigned		ref_total_;
# endif // WP_ENABLE_INFO
      };

    } // End of namespace internals.


    template <class InputIterator1,
	      class InputIterator2,
	      class Bottom = tool::bottom>
    struct fast_intersector:
      WP_INHERIT(public tool::iterable, fast_intersector<InputIterator1,
		 InputIterator2,
		 Bottom>)
    {
      fast_intersector(const InputIterator1& first1,
		       const InputIterator1& last1,
		       const InputIterator2& first2,
		       const InputIterator2& last2,
		       const unsigned		precision);

      void
	reset_lhs(const InputIterator1& first1,
		  const InputIterator1& last1);

      void
	reset_rhs(const InputIterator2& first2,
		  const InputIterator2& last2);


# ifdef WP_DONT_USE_HASH
      typedef std::set<typename InputIterator1::value_type>	arrivals_type;
# else
      typedef
	tool::hash_set<typename InputIterator1::value_type>	arrivals_type;

      enum
      {
	/**
	 * Size of the hash table that contains unique frames from
	 * the second trace.
	 *
	 * The worst case we have met has 7047619 unique frames.
	 * The closest power of 2 is 1 << 23 = 8388608.  To have an
	 * efficient hash table we must have a bucket count several
	 * orders of magnitude larger.  Thus 1 << 25 = (1 << 23) * 4.
	 */
	hash_bucket_count = 1 << 25
      };
# endif

    private:
      InputIterator2	first_;
      InputIterator2	last_;
      arrivals_type	arrivals_;
      unsigned	precision_;

      friend
	class internals::fast_intersector_iterator<InputIterator1,
	InputIterator2,
	Bottom,
	tool::bottom>;
    };

  } // End of namespace wpl::filter.

} // End of namespace wpl.

# include "fast_intersector.hxx"

#endif // ! FILTER_FAST_INTERSECTOR_HH_
