/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_FILE_ITERATOR_HH_
# define PCAP_FILE_ITERATOR_HH_

# include <istream>

# include <boost/scoped_ptr.hpp>

# include <wipal/pcap/pkt.hh>
# include <wipal/tool/valued_iterator.hh>

namespace wpl
{

  /*---------------------.
  | Forward declarations |
  `---------------------*/

  namespace pcap
  {
    /// WiPal pcap %internals - not designed for users.
    namespace internals
    {

      template <class, class>
      struct file_iterator;

    } // End of namespace wpl::pcap::internals.
  } // End of namespace wpl::pcap.


  /*--------------.
  | Virtual types |
  `--------------*/

  namespace tool
  {

    template <class S, class B>
    struct types< pcap::internals::file_iterator<S, B> >
    {
      typedef pkt::packet<S>	value_type;
      typedef S			iterable_type;
    };

  } // End of namespace wpl::tool.


  /*-------------.
  | Declarations |
  `-------------*/

  namespace pcap
  {

    namespace internals
    {

      /// Iterator over frames for @link istream %pcap istream's @endlink.
      template <class Stream, class Bottom = tool::bottom>
      struct file_iterator:
	public tool::valued_iterator<WP_GET_EXACT(Bottom,
						  file_iterator<Stream,
								Bottom>)>
      {
	/// @name Static hierarchies related types
	//@{
	typedef WP_GET_EXACT(Bottom, file_iterator<Stream,
						   Bottom>)	exact_type;
	typedef WP_TYPE(iterable_type, exact_type)		iterable_type;
	typedef WP_TYPE(value_type, exact_type)			value_type;
	//@}

	/// tool::iterable compatible constructor.
	file_iterator(const iterable_type& stream, bool end);

	/// Construct a frame iterator at a given %pcap %file position.
	file_iterator(const iterable_type&	stream,
		      const std::streampos&	pos,
		      unsigned			count,
		      std::istream* const	istream = 0);

	/// Copy constructor.
	file_iterator(const file_iterator&);

	/// Assignment operator.
	file_iterator&	operator = (const file_iterator&);

	/// @name Abstract method implementations
	//@{
	bool		equal(const file_iterator& rhs) const;
	void		increment();
	//@}

      protected:
	/// Retrieve the underlying C++ input stream.
	std::istream&	stream();

	/// Build a pkt::packet from a std::istream.
	void		increment_pcap();

      private:
	/// Link to the %stream this iterator comes from.
	const iterable_type*		stream_;

	/**
	 * @name Stream-related information
	 *
	 * If @c this->value() holds @c boost::none_t (i.e. the iterator
	 * is at the end), do not expect those fields to have
	 * a meaningful value!
	 */
	//@{

	/*
	 * Profiling showed std::ifstream's constructor takes much time,
	 * even when not requesting to open a file (which is a frequent
	 * case).  Thus we use a pointer instead to avoid that overhead.
	 * Of course, this implies the required allocation will slow
	 * down cases where we must open a file and initialize this
	 * attribute.
	 */
	boost::scoped_ptr<std::istream>	input_;
	std::streampos			pos_;	///<< Current file offset.
	unsigned			count_;	///<< Current packet id.

	//@}
      };

    } // End of namespace wpl::pcap::internals.

  } // End of namespace wpl::pcap.

} // End of namespace wpl.

# include "file_iterator.hxx"

#endif // ! PCAP_FILE_ITERATOR_HH_
