/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_LIST_HXX_
# define PCAP_LIST_HXX_

# include "list.hh"

namespace wpl
{

  namespace pcap
  {

    template <class Stream, class Bottom>
    list<Stream, Bottom>::list(const std::string& stream_names)
    {
      typedef std::vector<std::string>			stream_vect_t;
      typedef std::pair<stream_vect_t, std::string>	parse_res_t;

      const parse_res_t		pres = parse(stream_names);
      const stream_vect_t&	names = pres.first;

      std::string		summary;

      if (not pres.second.empty())
	meta_.add("addr", pres.second);
      switch (names.size())
	{
	case 0:
	  summary = "[]";
	  break;
	case 1:
	  summary = names.front();
	  break;
	default:
	  {
	    size_t	len = 0;
	    bool	cont = true;

	    while (cont and len <= names.front().size())
	      {
		const std::string	ref = names.front().substr(0, ++len);

		for (size_t i = 1; cont and i < names.size(); ++i)
		  if (names[i].substr(0, len) != ref)
		    cont = false;
	      }
	    summary = '[' + names.front().substr(0, len - 1) + "*]";
	    break;
	  }
	}
      meta_.add("name", summary);

      BOOST_FOREACH(const std::string& s, names)
	push_back(Stream (s));

      update_meta(this);
    }

    template <class Stream, class Bottom>
    list<Stream, Bottom>::list(const list& other):
      super_type (static_cast<const super_type&> (other)),
      meta_ (other.meta_)
    {
      update_meta(this);
    }

    template<class Stream, class Bottom>
    list<Stream, Bottom>&
    list<Stream, Bottom>::operator = (const list& rhs)
    {
      if (&rhs == this)
	return *this;

      super_type::operator = (rhs);
      meta_ = rhs.meta_;
      update_meta(this);
    }

    template <class Stream, class Bottom>
    std::ostream&
    list<Stream, Bottom>::print(std::ostream& os) const
    {
      return os << meta_["name"].get<std::string>();
    }

    template <class Stream, class Bottom>
    pkt::type
    list<Stream, Bottom>::type() const
    {
      if (this->empty())
	{
	  std::ostringstream os;

	  os << *this << "is empty";
	  throw std::invalid_argument (os.str());
	}
      return this->get(0).type();
    }

    template <class Stream, class Bottom>
    void
    list<Stream, Bottom>::expect(typename pkt::type t) const
    {
      for (size_t i = 0; i < this->size(); ++i)
	this->get(i).expect(t);
    }

    template <class Stream, class Bottom>
    const opt::list&
    list<Stream, Bottom>::meta() const
    {
      return meta_;
    }

    template <class Stream, class Bottom>
    std::pair<std::string, size_t>
    list<Stream, Bottom>::parse_word(const std::string&	   stream_names,
				     const size_t	   start,
				     const std::set<char>& stop_seps)
    {
      std::string	w;
      size_t		i;

      for (i = start;
	   i < stream_names.size() and not stop_seps.count(stream_names[i]);
	   ++i)
	{
	  if (stream_names[i] == escape_char)
	    ++i;
	  if (i < stream_names.size())
	    w.push_back(stream_names[i]);
	}
      return std::make_pair(w, i);
    }

    template <class Stream, class Bottom>
    std::pair<std::vector<std::string>, std::string>
    list<Stream, Bottom>::parse(const std::string& stream_names)
    {
      std::vector<std::string>	names;
      std::string		addr;
      std::set<char>		stop_sep_tname;

      stop_sep_tname.insert(trace_name_sep);
      stop_sep_tname.insert(addr_sep);

      size_t	i = 0;

      while (i < stream_names.size() and stream_names[i] != addr_sep)
	{
	  std::pair<std::string, size_t>  w = parse_word(stream_names, i,
							 stop_sep_tname);
	  names.push_back(w.first);
	  i = w.second;
	  if (i < stream_names.size() and stream_names[i] == trace_name_sep)
	    ++i;
	}
      if (i < stream_names.size() and stream_names[i] == addr_sep)
	addr = parse_word(stream_names, i + 1, std::set<char> ()).first;
      return std::make_pair(names, addr);
    }

    template <class Stream, class Bottom>
    void
    list<Stream, Bottom>::update_meta(list* p)
    {
      for (unsigned i = 0; i < this->size(); ++i)
	this->get(i).meta().add("list", p);
    }

    template <class S, class B>
    std::ostream&
    operator << (std::ostream& os, const list<S, B>& l)
    {
      return l.print(os);
    }

  } // End of namespace wpl::pcap.

} // End of namespace wpl.

#endif // ! PCAP_LIST_HXX_
