/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_OSTREAM_HXX_
# define PCAP_OSTREAM_HXX_

# include <wipal/pcap/base.hh>

# include "ostream.hh"

namespace wpl
{

  namespace pcap
  {

    inline
    ostream::ostream(const std::string&	filename,
		     const pkt::type	type,
		     const int32_t	zone,
		     const int32_t	snaplen):
      filename_ (filename),
      size_ (0)
    {
      setup(type, zone, snaplen);
    }

    template <class B>
    ostream::ostream(const std::string&	filename,
		     const stream<B>&	stream):
      filename_ (filename),
      size_ (0)
    {
      setup(stream.type(), stream.zone(), stream.snaplen());
    }

    template <class S>
    ostream::ostream(const std::string&	filename,
		     const list<S>&	list):
      filename_ (filename),
      size_ (0)
    {
      if (list.empty())
	{
	  std::ostringstream os;

	  os << list << " is empty";
	  throw std::invalid_argument (os.str());
	}

      const S& stream = list.get(0);

      setup(stream.type(), stream.zone(), stream.snaplen());
    }

    inline
    ostream&
    ostream::write(const pkt::metadata& meta, const void* bytes)
    {
      // PCAP header.
      {
	internals::file_frame_header h;

	h.ts.tv_sec	= meta.ts.tv_sec;
	h.ts.tv_usec	= meta.ts.tv_usec;
	h.caplen	= meta.caplen;
	h.len		= meta.len;

	if (not output_->write(reinterpret_cast<char*> (&h), sizeof (h)))
	  throw tool::write_error("Could not write PCAP packet header");
      }

      // Packet.
      if (not output_->write(static_cast<const char*> (bytes), meta.caplen))
	throw tool::write_error("Could not write packet data");

      if (++size_ % pcap::indexed_file<>::mark_step == 0)
	marks_.push_back(output_->tellp());

      return *this;
    }

    template <class S>
    ostream&
    ostream::operator << (const pkt::packet<S>& p)
    {
      return write(p.meta(), p.bytes());
    }

    template <class S>
    ostream&
    ostream::operator () (const pkt::packet<S>& p)
    {
      return operator << (p);
    }

    inline
    pcap::indexed_file<>
    ostream::file() const
    {
      return pcap::indexed_file<> (filename_, marks_, size_);
    }

    inline
    void
    ostream::setup(const pkt::type	type,
		   const int32_t	zone,
		   const int32_t	snaplen)
    {
      output_.reset(new std::ofstream (filename_.c_str(), (std::ios::out   |
							   std::ios::trunc |
							   std::ios::binary)));
      internals::file_header h;

      h.magic		= 0xa1b2c3d4;
      h.version_major	= internals::version_major;
      h.version_minor	= internals::version_minor;
      h.thiszone	= zone;
      h.sigfigs		= 0;
      h.snaplen		= snaplen;
      h.linktype	= type;

      if (not *output_)
	throw tool::file_error("Could not open " + filename_ + " for writing");
      if (not output_->write(reinterpret_cast<char*> (&h), sizeof (h)))
	throw tool::write_error(filename_ +
				": Could not write PCAP file header");

      marks_.push_back(output_->tellp());
    }

  } // End of namespace wpl::pcap.

} // End of namespace wpl.

#endif // ! PCAP_OSTREAM_HH_
