/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_PKT_HH_
# define PCAP_PKT_HH_

extern "C"
{
# include <time.h>
# include <inttypes.h>
}
# include <iostream>

# include <boost/shared_ptr.hpp>
# include <boost/array.hpp>

# include <wipal/tool/iterable.hh>
# include <wipal/tool/options.hh>

namespace wpl
{

  /// Packet-related compounds.
  namespace pkt
  {

    /**
     * Packet type.
     *
     * Protocol identifiers for packet.  Used by stream::type() to
     * know what data packet::bytes() returns.  Uses the same constants
     * as %pcap files.
     */
    enum type
      {
	LOOP			= 0,	///< BSD loopback encapsulation
	EN10MB			= 1,	///< Ethernet (10Mb)
	EN3MB			= 2,	///< Experimental Ethernet (3Mb)
	AX25			= 3,	///< Amateur Radio AX.25
	PRONET			= 4,	///< Proteon ProNET Token Ring
	CHAOS			= 5,	///< Chaos
	IEEE802			= 6,	///< IEEE 802 Networks
	ARCNET			= 7,	///< ARCNET, with BSD-style header
	SLIP			= 8,	///< Serial Line IP
	PPP			= 9,	///< Point-to-point Protocol
	FDDI			= 10,	///< FDDI
	IEEE802_11		= 105,	///< IEEE 802.11 wireless
	PRISM_HEADER		= 119,	///< Prism PHY-layer information
	IEEE802_11_RADIO	= 127,	///< Radiotap header
	IEEE802_11_RADIO_AVS	= 163	///< AVS header
      };

    /**
     * Meta-information for a @link packet network packet @endlink.
     *
     * A metadata object holds information regarding a frame/packet
     * that is not directly embedded inside the frame/packet's bytes.
     */
    struct metadata
    {
      metadata(unsigned	id,
	       unsigned	ts_sec,
	       unsigned ts_usec,
	       size_t	caplen,
	       size_t	len);

      bool operator == (const metadata&) const; // Does not check id.
      bool operator != (const metadata&) const;

      unsigned		id;	///< Packet index.
      struct timeval	ts;	///< Time of arrival.
      size_t		caplen;	///< Number of captured bytes.
      size_t		len;	///< Length of the frame/packet.
    };

    /**
     * Packet %stream.
     *
     * This is the base class for all packet streams,
     * e.g. pcap::stream or pcap::file.  A %packet %stream is an
     * iterable object.  You may iterate over packets like this:
     *
     * @code
     * // typedef typename pkt::stream<...>::iterator iterator;
     * // pkt::stream<...>& s = ...
     *
     * for (iterator i = s.begin(); i != s.end(); ++i)
     *   {
     *     // Access to i->meta() and i->bytes().
     *     //
     *     // *i is of type pkt::packet.
     *   }
     * @endcode
     */
    template <class Bottom>
    struct stream: WP_INHERIT(public tool::iterable, stream<Bottom>)
    {
      typedef WP_GET_EXACT(Bottom, stream<Bottom>)	exact_type;
      typedef tool::iterable<exact_type>		super_type;
      typedef WP_TYPE(iterator, exact_type)		iterator;
      typedef iterator					const_iterator;
      typedef typename iterator::value_type		packet;

      /**
       * Get the %stream's endianness w.r.t. the host system.
       *
       * Some %packet streams might be encoded with a custom
       * endianness by %packet sniffers.  This should not impact users
       * in theory because it is up to the %stream implementation to
       * decode it.  However sometimes the capture system's endianness
       * might affect %packet bytes (e.g. Prism headers) and nothing
       * can be done at the %stream implementation level to fix this.
       * Thus this function.
       */
      bool		swapped() const;

      /**
       * Get the %stream's link type.
       *
       * @see type.
       */
      pkt::type		type() const;

      /**
       * Throw an instance of @c std::invalid_argument if the
       * %stream's link type is different from @c t.
       */
      void		expect(pkt::type t) const;

      /// Print a human readable summary of the %stream.
      std::ostream&	print(std::ostream&) const;

      /// Default impletentation for print().
      std::ostream&	print_impl(std::ostream&) const;

      /**
       * Access to the stream metadata.
       *
       * @todo FIXME: Document.
       */
      //@{
      opt::list&	meta();
      const opt::list&	meta() const;
      //@}

    private:
      opt::list		meta_;
    };

    /// Print a human readable summary of a stream.
    template <class B>
    std::ostream&
    operator << (std::ostream& os, const stream<B>& s);

    /**
     * Network %packet.
     *
     * A packet holds both the %packet's @link metadata
     * meta-information @endlink and bytes.
     *
     * A %packet might optionally be attached to its source.  If
     * attached, the %packet has an index inside this source.
     *
     * Packets use shared pointers: you may copy them efficiently.
     */
    template <class Source>
    struct packet
    {
      /// Maximum length of a packet.
      enum { max_size = 65535 };

      /// C++ type of a packet's source.
      typedef Source		source_type;

      /// Constructors.
      //@{
      packet(unsigned		index,
	     unsigned		sec,
	     unsigned		usec,
	     size_t		caplen,
	     size_t		len = 0,
	     const Source*	src = 0);

      packet(const metadata&	meta,
	     const Source*	src = 0);
      //@}

      /// Get a pointer to the %packet source.  Return 0 if there is no source.
      const Source*	source_ptr() const;

      /// Get the %packet source.  Throw if the packet has no source.
      const Source&	source() const;

      /// Get whether this %packet come from an endianness-swapped trace.
      bool		swapped() const;

      /// Get the number of packet objects that share the bytes and metadata.
      long		use_count() const;

      /// Get whether other packet objects share the same bytes and metadata.
      bool		unique() const;

      /// @link metadata Meta-information @endlink accessors.
      //@{
      const metadata&	meta() const;
      metadata&		meta();
      //@}

      /// Access to %packet bytes.
      //@{
      const void*	bytes() const;
      void*		bytes();
      //@}

      /// Dump a summary of the %packet to a C++ stream.
      std::ostream&	print(std::ostream&) const;

      /// @name Equality checks
      //@{
      bool	operator == (const packet&) const;
      bool	operator != (const packet&) const;
      //@}

      /// Data embedded in a packet.
      struct shared_data
      {
	template <class B>
	shared_data(const stream<B>*	src,
		    const metadata&	meta);

	const Source*			src_;
	metadata			meta_;
	boost::array<uint8_t, max_size>	bytes_;
      };

    protected:
      boost::shared_ptr<shared_data>		data_;
    };

    /// Dump a summary of a packet to a C++ stream.
    template <class Source>
    std::ostream&
    operator << (std::ostream&, const packet<Source>&);

  } // End of namespace wpl::pkt.

} // End of namespace wpl.

# include "pkt.hxx"

#endif // ! PCAP_PKT_HH_
