/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_PKT_HXX_
# define PCAP_PKT_HXX_

# include <cstring>
# include <sstream>

# include "pkt.hh"

namespace wpl
{

  namespace pkt
  {

    /*--------------.
    | pkt::metadata |
    `--------------*/

    inline
    metadata::metadata(unsigned	id,
		       unsigned	ts_sec,
		       unsigned	ts_usec,
		       size_t	caplen,
		       size_t	len):
      id (id),
      caplen (caplen),
      len (len)
    {
      ts.tv_sec = ts_sec;
      ts.tv_usec = ts_usec;
    }

    inline
    bool
    metadata::operator == (const metadata& m) const
    {
      return (m.ts.tv_sec == ts.tv_sec and m.ts.tv_usec == ts.tv_usec and
	      m.caplen == caplen and m.len == len);
    }

    inline
    bool
    metadata::operator != (const metadata& m) const
    {
      return not operator == (m);
    }


    /*------------.
    | pkt::stream |
    `------------*/

    template <class B>
    bool
    stream<B>::swapped() const
    {
      return this->exact().swapped_impl();
    }

    template <class B>
    pkt::type
    stream<B>::type() const
    {
      return this->exact().type_impl();
    }

    template <class B>
    void
    stream<B>::expect(pkt::type t) const
    {
      if (type() != t)
	{
	  std::ostringstream os;

	  os << *this << ": Expected link type " << t << ", got " << type();
	  throw std::invalid_argument (os.str());
	}
    }

    template <class B>
    std::ostream&
    stream<B>::print(std::ostream& os) const
    {
      return this->exact().print_impl(os);
    }

    template <class B>
    std::ostream&
    stream<B>::print_impl(std::ostream& os) const
    {
      return os << typeid (this->exact()).name() << '@' << this;
    }

    template <class B>
    opt::list&
    stream<B>::meta()
    {
      return meta_;
    }

    template <class B>
    const opt::list&
    stream<B>::meta() const
    {
      return meta_;
    }

    template <class B>
    std::ostream&
    operator << (std::ostream& os, const stream<B>& s)
    {
      return s.print(os);
    }


    /*------------.
    | pkt::packet |
    `------------*/

    template <class Source>
    packet<Source>::packet(unsigned		index,
			   unsigned		sec,
			   unsigned		usec,
			   size_t		caplen,
			   size_t		len,
			   const Source*	src):
      data_ (new shared_data (src, metadata (index, sec, usec, caplen, len)))
    {
      if (not len and caplen)
	len = caplen;

      assert(caplen <= max_size);
    }

    template <class Source>
    packet<Source>::packet(const metadata&	meta,
			   const Source*	src):
      data_ (new shared_data (src, meta))
    {
    }

    template <class Source>
    const Source*
    packet<Source>::source_ptr() const
    {
      assert(data_);

      return data_->src_;
    }

    template <class Source>
    const Source&
    packet<Source>::source() const
    {
      assert(data_);

      if (not data_->src_)
	throw std::runtime_error ("Frame not attached to a source");
      return *data_->src_;
    }

    template <class Source>
    bool
    packet<Source>::swapped() const
    {
      const Source* const	src = source_ptr();

      return src ? src->swapped() : false;
    }

    template <class Source>
    long
    packet<Source>::use_count() const
    {
      return data_.use_count();
    }

    template <class Source>
    bool
    packet<Source>::unique() const
    {
      return data_.unique();
    }

    template <class Source>
    const metadata&
    packet<Source>::meta() const
    {
      assert(data_);

      return data_->meta_;
    }

    template <class Source>
    metadata&
    packet<Source>::meta()
    {
      assert(data_);

      return data_->meta_;
    }

    template <class Source>
    const void*
    packet<Source>::bytes() const
    {
      assert(data_);

      return data_->bytes_.begin();
    }

    template <class Source>
    void*
    packet<Source>::bytes()
    {
      assert(data_);

      return data_->bytes_.begin();
    }

    template <class Source>
    std::ostream&
    packet<Source>::print(std::ostream& os) const
    {
      const Source* const	src = source_ptr();

      if (src)
	os << *src << ':';
      return os << meta().id;
    }

    template <class Source>
    bool
    packet<Source>::
    operator == (const packet& rhs) const
    {
      return meta() == rhs.meta() and 0 == memcmp(bytes(), rhs.bytes(),
						  meta().caplen);
    }

    template <class Source>
    bool
    packet<Source>::
    operator != (const packet& rhs) const
    {
      return not (*this == rhs);
    }

    template <class Source>
    template <class B>
    packet<Source>::shared_data::shared_data(const stream<B>*	src,
					     const metadata&	meta):
      meta_ (meta),
      src_ (src->exact_ptr())
    {
    }

    template <class Source>
    std::ostream&
    operator << (std::ostream& os, const packet<Source>& p)
    {
      return p.print(os);
    }

  } // End of namespace wpl::pcap.

} // End of namespace wpl.

#endif // ! PCAP_PKT_HXX_
