/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_STREAM_HH_
# define PCAP_STREAM_HH_

extern "C"
{
# include <inttypes.h>
}
# include <cstdlib>
# include <string>
# include <vector>

# include <wipal/pcap/pkt.hh>
# include <wipal/pcap/file_iterator.hh>

namespace wpl
{

  /**
   * PCAP files input/output.
   *
   * This namespace holds the compounds required to perform PCAP (Packet
   * CAPture) files input/output.
   *
   * - In order to <b>read PCAP files</b>, one needs a @link descriptor
   *   PCAP descriptor @endlink object.  PCAP descriptors also enable
   *   users to instanciate dumper objects.
   * - In order to <b>write PCAP files</b>, one needs a dumper object.
   * - It is possible to <b>consider separate PCAP files as one unique
   *   trace</b> consisting of the concatenation of all files.  Do this
   *   with a list_of_traces object.
   * - Use a match object to <b>look for specific byte sequences</b>
   *   into @link pkt::packet packets @endlink.
   *
   * @see descriptor, dumper, list_of_traces, match, pkt::packet.
   */
  namespace pcap
  {

    template <class D>
    struct file;

    template <class D>
    struct indexed_file;

  } // End of namespace wpl::pcap.

  namespace tool
  {

    template <class B>
    struct types< pcap::file<B> >
    {
      typedef WP_GET_EXACT(B, pcap::file<B>)			exact_type;
      typedef pcap::internals::file_iterator<exact_type>	iterator;
    };

    template <class B>
    struct types< pcap::indexed_file<B> >
    {
      typedef WP_GET_EXACT(B, pcap::indexed_file<B>)		exact_type;
      typedef pcap::internals::file_iterator<exact_type>	iterator;
    };

  } // End of namespace wpl::tool.

  namespace pcap
  {

    /// %pcap %stream.
    template <class Bottom>
    struct stream: WP_INHERIT(public pkt::stream, stream<Bottom>)
    {
      /// Get the %stream's time adjustment in seconds (%pcap %zone).
      int32_t		zone() const;

      /// Get the %stream's snap length.
      int32_t		snaplen() const;
    };

    /// %pcap %stream based on std::istream.
    template <class Bottom>
    struct istream: WP_INHERIT(public stream, istream<Bottom>)
    {
      typedef WP_GET_EXACT(Bottom, istream<Bottom>)	exact_type;
      typedef stream<exact_type>			super_type;
      typedef WP_TYPE(iterator, exact_type)		iterator;
      typedef iterator					const_iterator;

      /// Take ownership of @c is and construct a pcap::istream.
      istream(std::istream* const is);

      std::ostream&	print_impl(std::ostream&) const;

      bool		swapped_impl() const;
      pkt::type		type_impl() const;
      int32_t		zone_impl() const;
      int32_t		snaplen_impl() const;

    private:
      bool		swap_;
      pkt::type		type_;
      int32_t		zone_;
      int32_t		snaplen_;
    };

    /// %pcap file.
    template <class Bottom = tool::bottom>
    struct file: WP_INHERIT(public istream, file<Bottom>)
    {
      typedef WP_GET_EXACT(Bottom, file<Bottom>)	exact_type;
      typedef istream<exact_type>			super_type;

      /// Construct an %istream from a file name.
      file(const std::string& filename);

      /**
       * Get a fresh std::ifstream object.
       *
       * If @c seek is true, the stream is positioned at the beginning
       * of packets.
       *
       * The stream needs too be delete'ed after use.
       */
      std::ifstream*		stream(bool seek = true) const;

      /// Get the %file size in bytes.
      std::streampos		filesize() const;

      /// Get the %file name.
      const std::string&	filename() const;

      std::ostream&			print_impl(std::ostream&) const;

    private:
      std::string	filename_;
      std::streampos	filesize_;
    };

    /**
     * %pcap %file featuring random packet access.
     *
     * @todo FIXME: Document random packet access.
     */
    template <class Bottom = tool::bottom>
    struct indexed_file: WP_INHERIT(public file, indexed_file<Bottom>)
    {
      typedef WP_GET_EXACT(Bottom, indexed_file<Bottom>)	exact_type;
      typedef file<exact_type>					super_type;
      typedef WP_TYPE(iterator, exact_type)			iterator;

      /**
       * Internal index type for random frame access.
       *
       * Only advanced users need to know about this.
       *
       * @see mark_step.
       */
      typedef std::vector<std::streampos>		index_type;

      enum
      {
	/**
	 * Register a mark every @c mark_step packets.
	 *
	 * A mark is an offset in the trace file.  Marks are
	 * stored in a vector at regular intervals, so fast
	 * access to random packets is possible.
	 *
	 * This is a trade-off: more marks allows faster access, but
	 * fewer marks uses less memory.
	 *
	 * Given a value of 4, with sizeof (std::streampos) == 8 and
	 * 100.000.000 packets in a file (which is quite a lot), this
	 * gives an index of about 200 MiB.
	 */
	mark_step	= 4,

	/**
	 * When loading traces, read packet data with steps of @c
	 * junk_len bytes.
	 *
	 * The more the better, but one must be careful not to grow
	 * the stack too much.  We use reads here instead of seeks
	 * not to break the underlying C++ buffering.  Tests showed
	 * this is faster.
	 */
	junk_len	= 4096
      };

      /// Construct a file object from a %file name.
      indexed_file(const std::string& filename);

      /**
       * Construct a file object from a %file name; use an already
       * constructed index.
       */
      indexed_file(const std::string&	filename,
		   const index_type&	marks,
		   unsigned		size);

      /**
       * Construct a file from a %file name, with progress indicator
       * callbacks.
       *
       * @param[in]	filename	The file's name.
       * @param[in]	pi_setup	Called before reading packets with
       *				two arguments.  The first one is
       *				the first packet's offset, the last
       *				one is the file size.
       * @param[in]	pi_update	Called periodically when reading
       *				packets.  The current offset in the
       *				file is passed as argument.  Returns
       *				either 0 (continue) or a pointer to
       *				a string (stop and throws
       *				@c interrupted(string) ).
       */
      template <typename ProgressInfoSetupType,
		typename ProgressInfoUpdateType>
      indexed_file(const std::string&		filename,
		   ProgressInfoSetupType&	pi_setup,
		   ProgressInfoUpdateType&	pi_update);

      /// Get an iterator pointing to a specific packet.
      iterator	operator [] (size_t);

      /// Get the number of packets in the %file.
      size_t	size() const;

    protected:
      void	warn(const std::streampos& pos, const std::string& msg);
      void	handle_truncation(const std::string& msg);

    private:
      template <typename ProgressInfoSetupType,
		typename ProgressInfoUpdateType>
      void
      setup(ProgressInfoSetupType&	pi_setup,
	    ProgressInfoUpdateType&	pi_update);

      index_type	marks_;
      size_t		size_;
    };

  } // End of namespace wpl::pcap.

} // End of namespace wpl.

# include "stream.hxx"

#endif // ! PCAP_STREAM_HH_
