/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_ADDR_HH_
# define TOOL_ADDR_HH_

# include <string>
# include <ostream>

namespace wpl
{

  namespace tool
  {

    /**
     * Generic network address.
     *
     * Protocol-agnostic network address, considered as a sequence of @c N
     * values of type @c ValueType.  For instance, IPv4 addresses are
     * 4 uint8_t, MAC addresses are 6 uint8_t, and IPv6 addresses are
     * 8 uint16_t.  @c HexNotation indicates whether the address' string
     * representation uses hexadecimal or decimal values.  @c SepNotation
     * is the character one should use to separate bytes in
     * the string representation.
     */
    template <class ValueType, unsigned N,
	      bool HexNotation, char SepNotation>
    struct addr
    {
      /// Construct a null address.
      addr();

      /**
       * Construct an address from a memory location.
       *
       * An alternative to using this constructor is to @c static_cast
       * the memory location:
       *
       * @code
       * typedef addr< // ...
       *               > addr_t;
       *
       * void* p = // To some memory containing an address.
       *
       * std::cout << *static_cast<addr_t*> (p) << std::endl;
       * @endcode
       */
      addr(const void* a);

      /**
       * Construct an address from a string representation.
       *
       * @e Example:
       *
       * @code
       * addr<uint8_t, 6, true,  ':'> mac (std::string ("00:16:ec:92:84:b8"));
       * addr<uint8_t, 4, false, '.'> ip4 (std::string ("192.168.1.1"));
       * @endcode
       */
      addr(const std::string&);

      /**
       * Construct an address from a stream (the address appears in the
       * stream as a string representation).
       *
       * @code
       * addr<uint8_t, 6, true, ':'> a (std::cin); // E.g. with 00:16:ec:92:84:b8
       *						 // on standard input.
       * @endcode
       */
      addr(std::istream&);

      /**
       * Set the current address according to a string representation
       * read from a stream.
       */
      std::istream&	read(std::istream&);

      /// @name Comparison operators
      //@{
      bool		operator < (const addr&) const;
      bool		operator == (const addr&) const;
      bool		operator != (const addr&) const;
      //@}

      /// @name Address' bytes access
      //@{
      ValueType&		operator [] (unsigned);
      const ValueType&	operator [] (unsigned) const;
      //@}

      /// Copy to a memory location.
      void			dump(void *) const;

      /// Get a null address (every byte is set to 0).
      static const addr&	null();

      /// Get a broadcast address (every byte is set to 255).
      static const addr&	broadcast();

      /**
       * Return the size of an address.
       *
       * Return the size of an address, i.e.:
       *
       * @code
       * N * sizeof (ValueType)
       * @endcode
       */
      size_t			size() const;

    private:

      ValueType	addr_[N];

    } __attribute__((__packed__));

    /// Dump an @link addr address @endlink in human readable notation.
    template <class VT, unsigned N, bool HN, char SN>
    std::ostream&
    operator << (std::ostream& os, const addr<VT, N, HN, SN>& a);

    /// Load an address from a stream (in string representation).
    template <class VT, unsigned N, bool HN, char SN>
    std::istream&
    operator >> (std::istream& is, addr<VT, N, HN, SN>& a);

    /// Make a human readable string from an @link addr address @endlink .
    template <class VT, unsigned N, bool HN, char SN>
    std::string
    make_string(const addr<VT, N, HN, SN>& a);

  } // End of namespace wpl::tool.

} // End of namespace wpl.

# include "addr.hxx"

#endif // ! TOOL_ADDR_HH_
