/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_ENDIANNESS_HXX_
# define TOOL_ENDIANNESS_HXX_

# include <cassert>

# include "endianness.hh"

namespace wpl
{

  namespace tool
  {

    namespace end
    {

      inline
      std::ostream&
      operator << (std::ostream& o, const endianness& e)
      {
	switch (e)
	  {
	  case big:
	    return o << "big endian";

	  case little:
	    return o << "little endian";

	  case raw:
	    return o << "same endianness as host";

	  case swap:
	    return o << "swap bytes";

	  case guess:
	    return o << "guess endianness according to data";

	  default:
	    assert(! "Internal error: undefined endianness.");
	    return o << "unkown endianness value";
	  }
      }

    } // End of namespace end.

    inline
    bool
    need_swap(end::endianness t, bool guess_swap)
    {
      static const int	i = 1;
      static const bool	l = reinterpret_cast<const char*> (&i)[0];

      switch (t)
	{
	case end::big:
	  return l;

	case end::little:
	  return not l;

	case end::raw:
	  return false;

	case end::swap:
	  return true;

	case end::guess:
	  return guess_swap;

	default:
	  assert(! "Internal error: undefined endianness.");
	  return false;
	}
    }


    /*----------------.
    | Unsigned shorts |
    `----------------*/

    inline
    uint16_t
    extract_unswapped_short_u(const void* p)
    {
      return *static_cast<const uint16_t*> (p);
    }

    inline
    uint16_t
    extract_swapped_short_u(const uint16_t s)
    {
      return (s >> 8) | (s << 8);
    }

    inline
    uint16_t
    extract_swapped_short_u(const void* p)
    {
      return extract_swapped_short_u(extract_unswapped_short_u(p));
    }

    inline
    uint16_t
    extract_short_u(const void* p, bool swapped)
    {
      return swapped ? extract_swapped_short_u(p) : extract_unswapped_short_u(p);
    }

    inline
    uint16_t
    extract_short_u(const uint16_t s, bool swapped)
    {
      return swapped ? extract_swapped_short_u(s) : s;
    }

    inline
    uint16_t
    extract_big_endian_short_u(const void *p)
    {
      const uint8_t *const c = static_cast<const uint8_t*> (p);

      return uint16_t (c[1]) | uint16_t (c[0]) << 8;
    }

    inline
    uint16_t
    extract_big_endian_short_u(const uint16_t s)
    {
      return extract_big_endian_short_u(&s);
    }

    inline
    uint16_t
    extract_little_endian_short_u(const void *p)
    {
      const uint8_t *const c = static_cast<const uint8_t*> (p);

      return uint16_t (c[0]) | uint16_t (c[1]) << 8;
    }

    inline
    uint16_t
    extract_little_endian_short_u(const uint16_t s)
    {
      return extract_little_endian_short_u(&s);
    }

    /*---------------.
    | Unsigned longs |
    `---------------*/

    inline
    uint32_t
    extract_unswapped_long_u(const void* p)
    {
      return *static_cast<const uint32_t*> (p);
    }

    inline
    uint32_t
    extract_swapped_long_u(const uint32_t l)
    {
      return ((l >> 24) | ((l & 0xFF0000) >> 8) |
	      ((l & 0xFF00) << 8) | (l << 24));
    }

    inline
    uint32_t
    extract_swapped_long_u(const void* p)
    {
      return extract_swapped_long_u(extract_unswapped_long_u(p));
    }

    inline
    uint32_t
    extract_long_u(const void* p, bool swapped)
    {
      return swapped ? extract_swapped_long_u(p) : extract_unswapped_long_u(p);
    }

    inline
    uint32_t
    extract_long_u(const uint32_t l, bool swapped)
    {
      return swapped ? extract_swapped_long_u(l) : l;
    }

    inline
    uint32_t
    extract_big_endian_long_u(const void* p)
    {
      const uint8_t *const c = static_cast<const uint8_t*> (p);

      return (c[0] << 24) | (c[1] << 16) | (c[2] << 8) | c[3];
    }

    inline
    uint32_t
    extract_big_endian_long_u(const uint32_t l)
    {
      return extract_big_endian_long_u(&l);
    }

    inline
    uint32_t
    extract_little_endian_long_u(const void* p)
    {
      const uint8_t *const c = static_cast<const uint8_t*> (p);

      return (c[3] << 24) | (c[2] << 16) | (c[1] << 8) | c[0];
    }

    inline
    uint32_t
    extract_little_endian_long_u(const uint32_t l)
    {
      return extract_little_endian_long_u(&l);
    }

    /*-------------.
    | Signed longs |
    `-------------*/

    inline
    int32_t
    extract_unswapped_long_s(const void* p)
    {
      return int32_t (extract_unswapped_long_u(p));
    }

    inline
    int32_t
    extract_swapped_long_s(const int32_t l)
    {
      return int32_t (extract_swapped_long_u(l));
    }

    inline
    int32_t
    extract_swapped_long_s(const void* p)
    {
      return int32_t (extract_swapped_long_u(p));
    }

    inline
    int32_t
    extract_long_s(const void* p, bool swapped)
    {
      return int32_t (extract_long_u(p, swapped));
    }

    inline
    int32_t
    extract_long_s(const int32_t l, bool swapped)
    {
      return int32_t (extract_long_u(l, swapped));
    }

    inline
    int32_t
    extract_big_endian_long_s(const void* p)
    {
      return int32_t (extract_big_endian_long_u(p));
    }

    inline
    int32_t
    extract_big_endian_long_s(const int32_t l)
    {
      return int32_t (extract_big_endian_long_u(l));
    }

    inline
    int32_t
    extract_little_endian_long_s(const void* p)
    {
      return int32_t (extract_little_endian_long_u(p));
    }

    inline
    int32_t
    extract_little_endian_long_s(const int32_t l)
    {
      return int32_t (extract_little_endian_long_u(l));
    }

  } // End of namespace wpl::tool.

} // End of namespace wpl.

#endif // ! TOOL_ENDIANNESS_HXX_
