/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_ITERATOR_HH_
# define TOOL_ITERATOR_HH_

# include <boost/optional.hpp>

# include <wipal/tool/static_base.hh>

namespace wpl
{

  namespace tool
  {

    template <class B>
    struct iterator;

    /// Virtual %types for iterator.  @see types
    template <class B>
    struct types< iterator<B> >
    {
      typedef undefined_type	value_type;
    };

    /**
     * Base interface for iterators.
     *
     * iterator provides the standard operators one expects from iterators
     * (@c == , @c != , @c ++ , @c * , @c -> ).
     *
     * To subclass iterator, you must provide the following methods:
     *
     * <ul>
     *
     *   <li>@c equal(const @c exact_type&) : Equality check.</li>
     *
     *   <li>@c increment() : Increment the iterator.</li>
     *
     *   <li>@c get() and @c get_ptr() : Indirection (as a reference and
     *   as a pointer).</li>
     *
     * </ul>
     */
    template <class Bottom>
    struct iterator: WP_INHERIT(public static_base, iterator<Bottom>)
    {
      /// Exact type for this template instantiation.
      typedef WP_GET_EXACT(Bottom, iterator<Bottom>)	exact_type;

      /// @name Iterator types.
      //@{
      typedef ssize_t					difference_type;
      typedef WP_TYPE(value_type, exact_type)		value_type;
      typedef value_type*				pointer;
      typedef value_type&				reference;
      typedef std::input_iterator_tag			iterator_category;
      //@}

      /// @name Equality checks
      //@{
      bool	operator == (const exact_type&) const;
      bool	operator != (const exact_type&) const;
      //@}

      /// @name Increment operators
      //@{
      const exact_type&	operator ++ ();
      exact_type	operator ++ (int);
      //@}

      /// @name Indirections
      //@{
      const value_type&	operator * () const;
      const value_type*	operator -> () const;
      //@}

    protected:
      /// Protected destructor: iterator is abstract.
      ~iterator();
    };

  } // End of namespace wpl::tool.

} // End of namespace wpl.

# include "iterator.hxx"

#endif // ! TOOL_ITERATOR_HH_
