/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_LINEAR_REGRESSION_HH_
# define TOOL_LINEAR_REGRESSION_HH_

# include <utility>

namespace wpl
{

  namespace tool
  {

    /// Linear regression adapters.
    namespace lr
    {

      /**
       * Linear regression adapter - make a @c std::pair suitable for
       * linear regression.
       *
       * For a given @c std::pair @c p, the x() method returns @c
       * p.first while the y() method returns @c p.second.
       *
       * @see linear_regression(), lr::pair_with_microseconds.
       */
      template <class ImplType>
      struct pair
      {
	/// Get @c p.first.
	template <class T1, class T2>
	ImplType
	x(const std::pair<T1, T2>& p) const;

	/// Get @c p.second.
	template <class T1, class T2>
	ImplType
	y(const std::pair<T1, T2>& p) const;
      };

      /**
       * Linear regression adapter - make a @c std::pair of elements
       * with %microseconds suitable for linear regression.
       *
       * For a given @c std::pair @c p, the x() method returns @c
       * p.first.microseconds() while the y() method returns @c
       * p.second.microseconds().
       *
       * @see linear_regression(), lr::pair.
       */
      template <class ImplType>
      struct pair_with_microseconds
      {
	/// Get @c p.first.microseconds().
	template <class T1, class T2>
	ImplType
	x(const std::pair<T1, T2>& p) const;

	/// Get @c p.second.microseconds().
	template <class T1, class T2>
	ImplType
	y(const std::pair<T1, T2>& p) const;
      };

    } // End of namespace tool::lr.

    /**
     * Perform a linear regression.
     *
     * Input data is provided by @c start and @c last.  @c adapt is
     * a function object containing both a @c x() and @c y() method
     * which respectively convert an instance of @c
     * InputIterator::value_type into its x and y components.
     *
     * @see lr::pair, lr::pair_with_microseconds.
     *
     * @return a pair of coefficients.
     */
    template <class ImplType,
	      class Adapter,
	      class InputIterator>
    std::pair<ImplType, ImplType>
    linear_regression(const InputIterator&	start,
		      const InputIterator&	last,
		      Adapter			adapt = Adapter ());

  } // End of namespace wpl::tool.

} // End of namespace wpl.

# include "linear_regression.hxx"

#endif // ! TOOL_LINEAR_REGRESSION_HH_
