/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_CTL_HH_
# define WIFI_FRAME_CTL_HH_

# include <string>

# include <wipal/wifi/addr.hh>

namespace wpl
{

  namespace wifi
  {

    /// IEEE 802.11 <i>control %frame</i> compounds.
    namespace ctl
    {

      /// IEEE 802.11 control %frame subtypes.
      namespace subtype
      {

	/// IEEE 802.11 control %frame subtypes.
	enum frame_subtype
	  {
	    ps_poll		= 10, ///< PS-Poll %frame identifier.
	    rts			= 11, ///< RTS %frame identifier.
	    cts			= 12, ///< CTS %frame identifier.
	    ack			= 13, ///< Ack. %frame identifier.
	    cf_end		= 14, ///< CF-End %frame identifier.
	    cf_end_cf_ack	= 15  ///< CF-End+CF-Ack %frame identifier.
	  };

	/// IEEE 802.11 control %frame subtype names.
	extern const std::string	names[16];

      } // End of namespace wpl::wifi::ctl::subtype.

      /// IEEE 802.11 <i>RTS %frame</i> compounds.
      namespace rts
      {

	/// RTS %frame structure.
	struct header
	{

	  /// RTS %frame address indices.
	  enum addr_index
	    {
	      ra	= 0,	///< Receiver address.
	      ta	= 1	///< Transmitter address.
	    };

	  uint16_t	frame_ctl;	///< Frame control field.
	  uint16_t	duration;	///< Frame duration field.
	  addr		addrs[2];	///< Frame addresses.
	} __attribute__((__packed__));
      } // End of namespace wpl::wifi::ctl::rts.

      /// IEEE 802.11 <i>CTS %frame</i> compounds.
      namespace cts
      {
	/// Ack. / CTS %frame structure.
	struct header
	{
	  uint16_t	frame_ctl;	///< Frame control field.
	  uint16_t	duration;	///< Frame duration field.
	  addr		ra;		///< Receiver address.
	} __attribute__((__packed__));
      } // End of namespace wpl::wifi::ctl::cts.

      /**
       * @namespace wpl::wifi::ctl::ack
       * IEEE 802.11 <i>acknowledgment %frame</i> compounds.
       */
      namespace ack = cts;

      /// IEEE 802.11 <i>PS-Poll %frame</i> compounds.
      namespace ps_poll
      {
	/// IEEE 802.11 <i>PS-Poll %frame</i> structure.
	struct header
	{

	  /// PS-Poll %frame address indices.
	  enum addr_index
	    {
	      bssid	= 0,	///< BSSID address.
	      ta	= 1	///< Transmitter address.
	    };

	  uint16_t	frame_ctl;	///< Frame control field.
	  uint16_t	aid;		///< Frame association identifier.
	  addr		addrs[2];	///< Frame addresses.
	} __attribute__((__packed__));
      } // End of namespace wpl::wifi::ctl::ps_poll.

      /// IEEE 802.11 <i>CF-End %frame</i> compounds.
      namespace cf_end
      {
	/// IEEE 802.11 <i>CF-End / CF-End+CF-Ack %frame</i> structure.
	struct header
	{

	  /// CF-End / CF-End+CF-Ack %frame address indices.
	  enum addr_index
	    {
	      ra	= 0,	///< Receiver address.
	      bssid	= 1	///< BSSID address.
	    };

	  uint16_t	frame_ctl;	///< Frame control field.
	  uint16_t	duration;	///< Frame duration field.
	  addr		addrs[2];	///< Frame addresses.
	} __attribute__((__packed__));
      } // End of namespace wpl::wifi::ctl::cf_end.

      /**
       * @namespace wpl::wifi::ctl::cf_end_cf_ack
       * IEEE 802.11 <i>CF-End+CF-Ack %frame</i> compounds.
       */
      namespace cf_end_cf_ack = cf_end;

    } // End of namespace wpl::wifi::ctl.

  } // End of namespace wpl::wifi.

} // End of namespace wpl.

#endif // ! WIFI_FRAME_CTL_HH_
