/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_ESSID_HXX_
# define WIFI_ESSID_HXX_

# include <sstream>
# include <iomanip>
# include <stdexcept>
# include <cctype>

# include <boost/foreach.hpp>

# include <wipal/wifi/essid.hh>

namespace wpl
{

  namespace wifi
  {

    inline
    essid::essid(const ssid_elt& se):
      std::string (se.ssid, se.ssid + se.length)
    {
    }

    inline
    essid::essid(const ssid_elt& se, size_t overrided_length):
      std::string (se.ssid, se.ssid + overrided_length)
    {
    }

    inline
    essid::essid(const std::string& s): std::string (s)
    {
    }

    inline
    std::ostream&
    essid::print(std::ostream& os) const
    {
      return os << escape();
    }

    inline
    std::istream&
    essid::read(std::istream& is, bool throw_on_parse_error)
    {
      clear();

      try
	{
	  if (is.flags() & std::istream::skipws)
	    skipws(is);
	  eat(is, '"');

	  char c = eat(is);

	  while (c != '"')
	    {
	      if (c == '\\' and (c = eat(is)) == 'x')
		{
		  unsigned char	v = 0;

		  v = xdigit(eat_xdigit(is));
		  c = eat(is);
		  if (std::isxdigit(c))
		    {
		      v = (v << 4) | xdigit(c);
		      c = eat(is);
		    }
		  push_back(v);
		  continue;
		}

	      push_back(c);
	      c = eat(is);
	    }
	}
      catch (...)
	{
	  is.setstate(is.rdstate() | std::istream::failbit);
	  if (throw_on_parse_error)
	    throw;
	}
      return is;
    }

    inline
    std::string
    essid::escape(size_t max) const
    {
      std::ostringstream	out;

      out << '"';
      BOOST_FOREACH(unsigned char c, *this)
	{
	  if (out.str().size() > max - 8) // worst case: \x2a..." (8 chars)
	    {
	      out << "...";
	      break;
	    }
	  if (c == '"')
	    out << "\\\"";
	  else if (c == '\\')
	    out << "\\\\";
	  else if (std::isprint(c))
	    out << c;
	  else
	    out << "\\x" << std::hex << std::setw (2) << std::setfill ('0')
		<< unsigned (c);
	}
      out << '"';

      return out.str();
    }

    inline
    void
    essid::skipws(std::istream& is)
    {
      while (std::isspace(is.peek()) and is and not is.eof())
	is.get();
    }

    inline
    unsigned char
    essid::eat(std::istream& is)
    {
      unsigned char r = is.get();

      if (not is or is.eof())
	throw
	  std::invalid_argument ("essid parse error: error or unexpected eof");
      return r;
    }

    inline
    unsigned char
    essid::eat_xdigit(std::istream& is)
    {
      unsigned char r = eat(is);

      if (not std::isxdigit(r))
	throw std::invalid_argument ("essid parse error: expected hex digit, "
				     "got '" + r + '\'');
      return r;
    }

    inline
    unsigned char
    essid::eat(std::istream& is, const unsigned char c)
    {
      unsigned char r = eat(is);

      if (r != c)
	{
	  std::ostringstream e;

	  e << "essid parse error: expected '" << c << "', got '" << r << '\'';
	  throw std::invalid_argument (e.str());
	}
      return r;
    }

    inline
    unsigned
    essid::xdigit(const unsigned char c)
    {
      if ('0' <= c and c <= '9')
	return c - '0';
      if ('a' <= c and c <= 'f')
	return c - 'a' + 10;
      if ('A' <= c and c <= 'F')
	return c - 'A' + 10;

      throw std::logic_error ("ESSID parse error: internal error, "
			      "fill a bug report");
    }

    inline
    std::ostream&
    operator << (std::ostream& o, const essid& s)
    {
      return s.print(o);
    }

    inline
    std::istream&
    operator >> (std::istream& i, essid& s)
    {
      return s.read(i);
    }

  } // End of namespace wpl::wifi.

} // End of namespace wpl.

#endif // ! WIFI_ESSID_HXX_
