/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FRAME_HH_
# define WIFI_FRAME_FRAME_HH_

# include <string>

extern "C"
{
# include <stdint.h>
}

# include <boost/optional.hpp>

# include <wipal/wifi/addr.hh>

namespace wpl
{

  namespace wifi
  {

    /// IEEE 802.11 %frame types.
    namespace type
    {

      /// IEEE 802.11 %frame types.
      enum frame_type
	{
	  management		= 0,	///< Management %frame identifier.
	  control		= 1,	///< Control %frame identifier.
	  data			= 2,	///< Data %frame identifier.
	  reserved		= 3	///< Reserved %frame %type.
	};

      /// IEEE 802.11 %frame type names.
      extern const std::string	names[4];

    } // End of namespace wpl::wifi::type.


    /// IEEE 802.11 %frame subtypes.
    namespace subtype
    {

      /// IEEE 802.11 %frame subtype names (2-dimensional array).
      extern const std::string* const	names[4];

    } // End of namespace wpl::wifi::subtype.

    /*---------------.
    | Control fields |
    `---------------*/

    /// Get a frame's <i>protocol version</i>.
    unsigned
    protocol_version_of(const void*);

    /// Get a frame's @e %type.
    type::frame_type
    type_of(const void*);

    /// Get a frame's @e subtype.
    unsigned
    subtype_of(const void*);

    /// Get a frame's <i>control flags</i>.
    uint8_t
    control_flags_of(const void*);

    /*--------------.
    | Control flags |
    `--------------*/

    /// Extract a specific flag from a frame's <i>control flags</i>.
    bool
    control_flag(const unsigned flag, const void* frame);

    /// Get a frame's <i>to distribution system</i> flag.
    bool
    to_ds(const void*);

    /// Get a frame's <i>from distribution system</i> flag.
    bool
    from_ds(const void*);

    /*---------.
    | Duration |
    `---------*/

    /// Get a frame's @e duration.
    unsigned
    duration_of(const void*);

    /*-----------------.
    | Sequence control |
    `-----------------*/

    boost::optional<unsigned>
    fragment_number_of(const void*, size_t caplen);

    boost::optional<unsigned>
    sequence_number_of(const void*, size_t caplen);

    /*------------------.
    | Misc. information |
    `------------------*/

    /**
     * Get a pointer to the  frame's transmitter address.
     *
     * When the %frame holds the address of its transmitter, return
     * a pointer to it.  Return 0 else.  Note that this function
     * extracts the <i>transmitter address</i>, not the <i>sender
     * address</i>.
     */
    const addr*
    transmitter_address(const void* frame);

    /**
     * Get a pointer to the frame's @c BSSID address.
     *
     * When the %frame holds an address that represents a @c BSSID
     * address, return a pointer to it.  Return 0 else.
     *
     * Some frames holds multiple @c BSSID addresses (%data frames with
     * both the <i>To DS</i> and <i>From DS</i> flags set).  In this
     * case, return the <i>transmitter @c BSSID address</i>.
     */
    const addr*
    bssid_address(const void*, size_t caplen = 30);

  } // End of namespace wpl::wifi.

} // End of namespace wpl.

# include "frame.hxx"

#endif // ! WIFI_FRAME_FRAME_HH_
