/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_STATS_ACTIVITY_HXX_
# define WIFI_STATS_ACTIVITY_HXX_

# include <cstring>
# include <cassert>

# include <boost/foreach.hpp>

# include <wipal/wifi/stats/activity.hh>

namespace wpl
{

  namespace wifi
  {

    namespace stats
    {

      inline
      activity::activity(const void*)
      {
      }

      template <class Frame>
      void
      activity::account_frame(const Frame&		frame,
			      const size_t&		len80211,
			      const type::frame_type&	type,
			      const unsigned&		subtype,
			      const bool&		,
			      const addr* const&	,
			      const addr* const&	,
			      const bool&		from_ap)
      {
	try
	  {
	    const int k = tracker_.update(frame.microseconds());

	    if (k > 1)
	      std::cerr << frame << ": [activity] warning: was inactive for "
			<< (k * epoch_length / 1000000) << " seconds (i.e. "
			<< k << " epochs)." << std::endl;
	  }
	catch (const non_increasing_timestamp&)
	  {
	    std::cerr << frame << ": [activity] Decreasing timestamp, "
		      << "ignoring." << std::endl;
	    return; // We are not able to account this frame, ignore it.
	  }

	epoch&	e = tracker_.current();

	++e.count;
	++e.type_count[type];
	++e.subtype_count[type][subtype];
	e.size += len80211;
	e.type_size[type] += len80211;
	e.subtype_size[type][subtype] += len80211;
	if (from_ap)
	  {
	    ++e.ap_count;
	    e.ap_size += len80211;
	  }
      }

      inline
      std::ostream&
      activity::print(std::ostream& o) const
      {
	o << "begin activity\n";

	BOOST_FOREACH(const epoch& e, tracker_.epochs())
	  o << "  "
	    << e.count << ' '
	    << e.size << ' '
	    << e.type_size[type::management] << ' '
	    << e.type_size[type::data] << ' '
	    << e.ap_size << '\n';
	return o << "end activity" << std::endl;
      }

      inline
      void
      activity::restart()
      {
	tracker_.restart();
      }

      inline
      activity::epoch::epoch(): count (0), ap_count (0), size (0), ap_size (0)
      {
	memset(type_count, 0, sizeof (type_count));
	memset(type_size, 0, sizeof (type_size));
	memset(subtype_count, 0, sizeof (subtype_count));
	memset(subtype_size, 0, sizeof (subtype_size));
      }

    } // End of namespace wpl::wifi::stats.

  } // End of namespace wpl::wifi.

} // End of namespace wpl.

#endif // ! WIFI_STATS_ACTIVITY_HXX_
