/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_STATS_EPOCH_TRACKER_HXX_
# define WIFI_STATS_EPOCH_TRACKER_HXX_

# include <cassert>

# include <wipal/wifi/stats/epoch_tracker.hh>

namespace wpl
{

  namespace wifi
  {

    namespace stats
    {

      template <class T, unsigned EL>
      epoch_tracker<T, EL>::epoch_tracker():
	epochs_ (1),
	epoch_idx_ (0)
      {
      }

      template <class T, unsigned EL>
      void
      epoch_tracker<T, EL>::restart()
      {
	epoch_start_ = boost::none_t ();
	epoch_idx_ = 0;
	assert(not epochs_.empty());
      }

      template <class T, unsigned EpochLength>
      int
      epoch_tracker<T, EpochLength>::
      update(const tool::microseconds& timestamp)
      {
	if (not epoch_start_)
	  {
	    assert(not epochs_.empty());
	    assert(0 == epoch_idx_);

	    epoch_start_ = timestamp;
	    return 1;
	  }

	tool::microseconds&		epoch_start = epoch_start_.get();
	const tool::microseconds	dt = timestamp - epoch_start;

	if (dt < 0)
	  throw non_increasing_timestamp ("Timestamp located before "
					  "current epoch");
	if (dt > EpochLength)
	  {
	    const tool::microseconds	e = EpochLength;
	    const unsigned		k = dt.get_div_by(EpochLength);

	    epoch_start += e * k;
	    epoch_idx_ += k;
	    if (epoch_idx_ >= epochs_.size())
	      epochs_.resize(epoch_idx_ + 1);
	    return int (k);
	  }
	return 0;
      }

      template <class T, unsigned EL>
      const std::vector<T>&
      epoch_tracker<T, EL>::epochs() const
      {
	return epochs_;
      }

      template <class T, unsigned EL>
      std::vector<T>&
      epoch_tracker<T, EL>::epochs()
      {
	return epochs_;
      }

      template <class T, unsigned EL>
      const T&
      epoch_tracker<T, EL>::current() const
      {
	return epochs_[epoch_idx_];
      }

      template <class T, unsigned EL>
      T&
      epoch_tracker<T, EL>::current()
      {
	return epochs_[epoch_idx_];
      }

      inline
      non_increasing_timestamp::
      non_increasing_timestamp(const std::string& what):
	std::invalid_argument (what)
      {
      }

    } // End of namespace wpl::wifi::stats.

  } // End of namespace wpl::wifi.

} // End of namespace wpl.

#endif // ! WIFI_STATS_EPOCH_TRACKER_HXX_
