/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_STATS_MISSED_ESTIMATIONS_HXX_
# define WIFI_STATS_MISSED_ESTIMATIONS_HXX_

# include <wipal/wifi/stats/missed_estimations.hh>
# include <wipal/wifi/frame.hh>
# include <wipal/wifi/mgt.hh>

namespace wpl
{

  namespace wifi
  {

    namespace stats
    {

      template <class H, class T>
      missed_estimations::missed_estimations(const module_list<H, T>* ml):
	missed_beacons_ (0),
	missed_ (0),
	sc_ (&ml->template get<simple_counters>())
      {
      }

      template <class Frame>
      void
      missed_estimations::
      account_beacon(const Frame&		frame,
		     const addr* const&		ap,
		     const addr* const&		,
		     const optional_string&	,
		     const bool&		)
      {
	if (not ap)
	  return;

	const tool::microseconds&	stamp = frame.microseconds();
	beacons_map::iterator		i = beacons_.find(*ap);

	if (i == beacons_.end())
	  {
	    beacons_.insert(std::make_pair(*ap, stamp));
	    return;
	  }

	const tool::microseconds	dt = stamp - i->second;

	if (dt < 0)
	  std::cerr << frame << ": Time-travelling beacon!" << std::endl;
	else
	  {
	    const unsigned		delta = dt.get_div_by(beacon_interval);

	    /* Due to timers imprecision It is usual delta is zero.
	     * e.g.  if (tstamp - i->second) is 102396 or 101299 and
	     * beacon_interval is 102400(those are real empirical
	     * values).
	     */
	    if (1 < delta)
	      missed_beacons_ += delta - 1;
	  }

	i->second = stamp;
      }

      inline
      void
      missed_estimations::account_gap(const unsigned& gap_length)
      {
	missed_ += gap_length;
      }

      inline
      void
      missed_estimations::account_miss()
      {
	++missed_;
      }

      inline
      std::ostream&
      missed_estimations::print(std::ostream& o) const
      {
	const unsigned	       stc = sc_->subtype_count(type::management,
							mgt::subtype::beacon);

	o << "estimated missed beacons: " << missed_beacons_ << '\n'
	  << "estimated missed beacons ratio: "
	  <<   (double (missed_beacons_) / (missed_beacons_ + stc)) << '\n'
	  << "estimated missed frames: " << missed_ << '\n'
	  << "estimated missed frames ratio: "
	  <<   (double (missed_) / (missed_ + sc_->total())) << std::endl;
	return o;
      }

    } // End of namespace wpl::wifi::stats.

  } // End of namespace wpl::wifi.

} // End of namespace wpl.

#endif // ! WIFI_STATS_MISSED_ESTIMATIONS_HXX_
