/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_STATS_STATS_HH_
# define WIFI_STATS_STATS_HH_

extern "C"
{
# include <stdint.h>
}

# include <map>
# include <utility>

# include <boost/tuple/tuple.hpp>
# include <boost/tuple/tuple_comparison.hpp>
# include <boost/optional.hpp>
# include <boost/preprocessor/tuple/elem.hpp>
# include <boost/preprocessor/seq/fold_left.hpp>
# include <boost/preprocessor/list/enum.hpp>

# include <wipal/wifi/stats/delimiter_frames.hh>
# include <wipal/wifi/stats/simple_counters.hh>
# include <wipal/wifi/stats/gap_lengths.hh>
# include <wipal/wifi/stats/missed_estimations.hh>
# include <wipal/wifi/stats/tfi_plot.hh>
# include <wipal/wifi/stats/bss_stats.hh>
# include <wipal/wifi/stats/ssid_stats.hh>
# include <wipal/wifi/stats/activity.hh>
# include <wipal/wifi/stats/growth.hh>
# include <wipal/wifi/stats/on_off_data.hh>
# include <wipal/wifi/stats/sta_counters.hh>
# include <wipal/wifi/stats/gaps.hh>
# include <wipal/tool/microseconds.hh>
# include <wipal/wifi/addr.hh>
# include <wipal/wifi/essid.hh>
# include <wipal/wifi/seqctl.hh>
# include <wipal/wifi/dissector/default_hooks.hh>
# include <wipal/wifi/mgt.hh>

namespace wpl
{

  namespace wifi
  {

    namespace stats
    {

      struct stats: public abstract_stats
      {
	template <class PhyHeader, class Frame>
	void
	account(const Frame& f);

	template <class Module>
	const Module&
	get() const;

	std::ostream&			print(std::ostream& o) const;
	void				restart();

      private: // Type declarations and enums.

	struct state
	{
	  enum
	    {
	      lifetime = 60 * 1000 * 1000, // 60 second.
	      min_frame_sending_time = 106
	    };

	  state(unsigned			last_frame_id,
		const seqctl&			last_sc,
		const tool::microseconds&	last_rx_stamp);

	  unsigned		last_frame_id;
	  seqctl		last_sc;
	  tool::microseconds	last_rx_stamp;
	};

	typedef std::map<seqnum_counter_id, state>	state_map;

	struct hooks: public dissector_default_hooks
	{
	  hooks();

	  void	invalid_type_or_subtype_hook(const void*, size_t,
					     dissector_status::status);
	  void	addr_hook(const void*, size_t, unsigned, const addr&);
	  void	seq_ctl_hook(const void*, size_t, unsigned, unsigned);
	  void	qos_ctl_hook(const void*, size_t,
			     unsigned, bool, unsigned, unsigned);
	  void	beacon_hook(const mgt::header*, size_t);

	  struct beacon_data
	  {
	    beacon_data(const mgt::header* header, size_t caplen);

	    const addr*				ap;
	    const addr*				bss;
	    bool				ibss;
	    boost::optional<std::string>	ssid;
	    // FIXME: TODO: Previous line: change for boost::optional<essid>.
	  };

	  typedef boost::optional<beacon_data>		optional_beacon_data;
	  typedef boost::optional<seqctl>		optional_seqctl;
	  typedef boost::optional<seqnum_counter_id>	optional_seq_counter_id;

	  type::frame_type				type;
	  unsigned					subtype;
	  bool						retx;
	  const addr*					a1;

	  // NULL if not present.
	  const addr*					a2;

	  // List of real addr., no BSSID.
	  std::vector< std::pair<int, const addr*> >	addrs;

	  optional_seqctl				sc;
	  optional_seq_counter_id			id;
	  optional_beacon_data				beacon;
	};

	enum { growths_epoch = 60 * 1000 * 1000 }; // 1 minute.

# define WP_stats_modules						\
	(sta_counters)(on_off_data)(beacon_items_growth<growths_epoch>)	\
	(addr_growth<growths_epoch>)					\
	(activity)(ssid_stats)(bss_stats)(tfi_plot)(gap_lengths)	\
	(simple_counters)(missed_estimations)(delimiter_frames) /* (gaps) */


	// About the following preprocessor frenzy: this is just some
	// magic to convert WP_stats_modules into a proper static
	// module_list<> declaration.

# define WP_op(U, State, Elem)						     \
	((module_list<Elem, BOOST_PP_TUPLE_ELEM(2, 0, State)),		     \
	 BOOST_PP_TUPLE_ELEM(2, 1, State)>)
# define WP_tuple							     \
	BOOST_PP_SEQ_FOLD_LEFT(WP_op, ((BOOST_PP_SEQ_HEAD(WP_stats_modules), \
					BOOST_PP_NIL),),		     \
			       BOOST_PP_SEQ_TAIL(WP_stats_modules))
# define WP_module_list_type						     \
	BOOST_PP_LIST_ENUM(BOOST_PP_TUPLE_ELEM(2, 0, WP_tuple))		     \
	BOOST_PP_TUPLE_ELEM(2, 1, WP_tuple)

	typedef WP_module_list_type module_list_type;

# undef WP_module_list_type
# undef WP_tuple
# undef WP_op
# undef WP_stats_modules


      private: // Methods.

	template <class Frame>
	void
	update_states(const Frame& f, const dissector<hooks>& d);

      private: // Attributes.
	state_map		states_;
	std::set<addr>		aps_;
	module_list_type	modules_;
      };

    } // End of namespace wpl::wifi::stats.

  } // End of namespace wpl::wifi.

} // End of namespace wpl.

# include "stats.hxx"

#endif // ! WIFI_STATS_STATS_HH_
