/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <stdexcept>
#include <iostream>
#include <cstdlib>

#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include <gmpxx.h>

#include <wipal/tool/args.hh>
#include <wipal/pcap/ostream.hh>
#include <wipal/filter/time_adjuster.hh>
#include <wipal/filter/fast_intersector.hh>
#include <wipal/pcap/list.hh>
#include <wipal/wifi/addr_mapping.hh>

namespace
{

  template <class T>
  struct sync_frames_dumper
  {
    sync_frames_dumper(wpl::pcap::ostream& dump): dump_ (dump)
    {
    }

    template <class TimeAdjuster>
    void	operator () (const TimeAdjuster& a)
    {
      a.for_each(dump_);
    }

  private:
    wpl::pcap::ostream&	dump_;
  };

  struct main_helper
  {
    template <class UniqueId>
    int
    operator () (wpl::opt::list& options) const
    {
      using namespace wpl;
      using filter::provide_time_adjuster;

      mpf_set_default_prec(WP_GMP_FLOAT_PRECISION);

      const args::input&	input = opt::get<args::input>(options["proc"]
							      ["input"]);
      pcap::list<>	traces1 (input[0]);
      pcap::list<>	traces2 (input[1]);

      const pkt::type	t = traces1.type();

      args::check_options(options, t);
      traces1.expect(t);
      traces2.expect(t);

      const args::output&	output = opt::get<args::output>(options["proc"]
								["output"]);
      pcap::ostream				os (output[0], traces2);
      sync_frames_dumper< pcap::list<> >	dump (os);

      switch (t)
	{
#define WP_case(R, Unused, LT)						\
	case pkt::BOOST_PP_TUPLE_ELEM(2, 0, LT):			\
	  provide_time_adjuster<UniqueId, BOOST_PP_TUPLE_ELEM(2, 1, LT), \
				filter::fast_intersector>		\
	    (traces1.begin(), traces1.end(),				\
	     traces2.begin(), traces2.end(),				\
	     opt::get<opt::list>(options["internals"]), dump);		\
	  break;

	  BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

#undef WP_case

	default:
	  args::unsupported_linktype(t);
	}
      return EXIT_SUCCESS;
    }

  }; // End of struct main_helper.

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      using namespace wpl;
      namespace def = args::def;

      opt::list	opts ((opt::name ("internals") -= opt::list
		       ((opt::name ("endianness") -= def::endianness(),
			 opt::name ("ref_blacklist") -= def::ref_blacklist(),
			 opt::name ("ignore_noisy_prism") -=
						     def::ignore_noisy_prism(),
			 opt::name ("mapping") -= def::mapping(),
			 opt::name ("precision") -= def::precision(),
			 opt::name ("phy_tstamps") -= def::phy_tstamps())),

		       opt::name ("misc") -= opt::list
		       ((opt::name ("mac_map") -= def::mac_map(),
			 opt::name ("attributes") -= def::attributes()))));

      args::parse(argc, argv, opts, args::count (2), args::count (1));
      return args::dispatch_unique_id(opts, main_helper ());
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
