import cairo

class Geometry:

    def __init__(self, traces_max):

        def __pt(mm):
            return mm * 1000 / 352.8

        self.__Width    = __pt(210)
        self.__Height   = __pt(297)

        self.__LMargin = __pt(10) # Left
        self.__RMargin = __pt(10) # Right
        self.__TMargin = __pt(10) # Top
        self.__BMargin = __pt(10) # Bottom

        self.__ThinLWidth  = 0.5
        self.__ThickLWidth = 3

        self.__TraceWidth       = __pt(10)
        self.__TraceHeight      = __pt(3.07)
        self.__TraceFontSize    = __pt(2.5)
        self.__TraceBaseLine    = (self.__TraceHeight - \
                                       self.__TraceFontSize) / 2
        self.__TraceRIndent     = __pt(1)
        self.__TraceFont        = "Sans"
        self.__TraceFontSlant   = cairo.FONT_SLANT_NORMAL
        self.__TraceFontWeight  = cairo.FONT_WEIGHT_NORMAL

        self.__AddrWidth        = __pt(30)
        self.__AddrHeight       = self.__TraceHeight * traces_max
        self.__AddrFontSize     = __pt(2.5)
        self.__AddrIndent       = __pt(1.5)
        self.__AddrBaseLine = \
            float (self.__AddrHeight - self.__AddrFontSize) / 2
        self.__AddrFont         = "Monospace"
        self.__AddrFontSlant    = cairo.FONT_SLANT_NORMAL
        self.__AddrFontWeight   = cairo.FONT_WEIGHT_BOLD

        self.__EvLMargin        = __pt(3)
        self.__EvRMargin        = __pt(2)
        self.__EvBaseLine       = float (self.__TraceHeight) / 2
        self.__EvSize           = self.__ThickLWidth + 0.5

        self.__EvWidth = self.__Width - self.__LMargin - self.__RMargin - \
            self.__AddrWidth - self.__TraceWidth - \
            self.__EvLMargin - self.__EvRMargin

        self.__AddrsPerPage      = \
            int ((self.__Height - self.__TMargin - self.__BMargin) / \
                     self.__AddrHeight)


    def set_thin_line_width(self, ctx):
        ctx.set_line_width(self.__ThinLWidth)

    def set_thick_line_width(self, ctx):
        ctx.set_line_width(self.__ThickLWidth)

    def ev_size(self):
        return self.__EvSize

    def addrs_per_page(self):
        return self.__AddrsPerPage

    def vline_ys(self):
        return (self.__TMargin,
                self.__TMargin + self.__AddrHeight * self.__AddrsPerPage)


    def width(self):
        return self.__Width

    def height(self):
        return self.__Height


    def addr_row_left(self, i):
        k = i % self.__AddrsPerPage
        return (self.__LMargin, self.__TMargin + (k + 1) * self.__AddrHeight)

    def addr_row_right(self, i):
        x, y = self.addr_row_left(i)
        return (x + self.__AddrWidth, y)

    def addr_text(self, i):
        x, y = self.addr_row_left(i)
        return (x + self.__AddrIndent, y - self.__AddrBaseLine)

    def set_addr_font(self, ctx):
        ctx.select_font_face(self.__AddrFont, self.__AddrFontSlant,
                             self.__AddrFontWeight)
        ctx.set_font_size(self.__AddrFontSize)


    def trace_row_left(self, i, j):
        x, y = self.addr_row_right(i)
        return x, y - (j * self.__TraceHeight)

    def trace_row_right(self, i, j):
        x, y = self.trace_row_left(i, j)
        return x + self.__TraceWidth, y

    def trace_text(self, i, j, text_width):
        x, y = self.trace_row_right(i, j)
        return (x - self.__TraceRIndent - text_width, y - self.__TraceBaseLine)

    def set_trace_font(self, ctx):
        ctx.select_font_face(self.__TraceFont, self.__TraceFontSlant,
                             self.__TraceFontWeight)
        ctx.set_font_size(self.__TraceFontSize)


    def ev_row_left(self, i, j):
        return self.trace_row_right(i, j)

    def ev_row_right(self, i, j):
        x, y = self.ev_row_left(i, j)
        return (self.__Width - self.__RMargin, y)

    def ev(self, i, j, t):
        x, y = self.ev_row_left(i, j)
        offset = t * self.__EvWidth
        return x + self.__EvLMargin + offset, y - self.__EvBaseLine

class Plot:

    def __line(self, (x1, y1), (x2, y2)):
        c = self.__ctx
        c.move_to(x1, y1)
        c.line_to(x2, y2)
        c.stroke()

    def __text(self, (x, y), text):
        c = self.__ctx
        c.move_to(x, y)
        c.show_text(text)
        c.stroke()

    def __bar(self, (x, y), size):
        c = self.__ctx
        c.move_to(x, y - size)
        c.line_to(x, y + size)
        c.stroke()

    def __cross(self, (x, y), size):
        c = self.__ctx
        tl = x - size, y - size
        tr = x + size, y - size
        bl = x - size, y + size
        br = x + size, y + size

        c.move_to(tl[0], tl[1])
        c.line_to(br[0], br[1])
        c.move_to(tr[0], tr[1])
        c.line_to(bl[0], bl[1])
        c.stroke()

    def __init__(self, geo, surf, ticstep):
        self.__BeginColor       =  (0.00, 0.50, 0.00)
        self.__EndColor         =  (0.50, 0.00, 0.00)
        self.__AddrColor        =  (0.00, 0.00, 0.00)
        self.__TraceColors      = ((1.00, 0.00, 0.00),
                                   (0.00, 0.75, 0.00),
                                   (0.00, 0.00, 1.00),
                                   (0.75, 0.75, 0.00),
                                   (0.00, 1.00, 1.00),
                                   (1.00, 0.00, 1.00))
        self.__FrameColor       =  (0.00, 0.00, 0.00)
        self.__ShadowColor      =  (0.85, 0.85, 0.85)

        self.__TicStep = ticstep

        self.__ctx = cairo.Context (surf)
        self.__geo = geo
        self.__n   = 0


    def addr(self, i, text):
        c, g = self.__ctx, self.__geo

        if self.__n % g.addrs_per_page() == 0:
            if self.__n:
                y1, y2 = g.vline_ys()
                x = g.trace_row_right(0, 0)[0]
                g.set_thin_line_width(c)
                c.set_source_rgb(self.__FrameColor[0], self.__FrameColor[1],
                                 self.__FrameColor[2])
                self.__line((x, y1), (x, y2))
                self.__text((g.ev_row_right(0, 0)[0], y2),
                            str(self.__n / g.addrs_per_page()))
                c.show_page()
        else:
            g.set_thin_line_width(c)
            c.set_source_rgb(self.__FrameColor[0], self.__FrameColor[1],
                             self.__FrameColor[2])
            self.__line(g.addr_row_left(i - 1), g.ev_row_right(i - 1, 0))
        g.set_addr_font(c)
        c.set_source_rgb(self.__AddrColor[0], self.__AddrColor[1],
                         self.__AddrColor[2])
        self.__text(g.addr_text(i), text)
        self.__n = self.__n + 1

    def trace(self, i, j, text):
        c, g = self.__ctx, self.__geo

        g.set_thin_line_width(c)
        c.set_source_rgb(self.__ShadowColor[0], self.__ShadowColor[1],
                         self.__ShadowColor[2])
        self.__line(g.ev(i, j, 0), g.ev(i, j, 1))
        t = 0.0
        while t <= 1.0:
            self.__bar(g.ev(i, j, t), g.ev_size())
            t += self.__TicStep

        g.set_trace_font(c)
        c.set_source_rgb(self.__TraceColors[j][0], self.__TraceColors[j][1],
                         self.__TraceColors[j][2])
        self.__text(g.trace_text(i, j, c.text_extents(text)[2]), text)

    def event(self, i, j, ev, old_ev):
        c, g = self.__ctx, self.__geo
        p = g.ev(i, j, ev[0])

        if (ev[1]):
            g.set_thin_line_width(c)
            c.set_source_rgb(self.__BeginColor[0], self.__BeginColor[1],
                             self.__BeginColor[2])
            self.__bar(p, g.ev_size())
        else:
            if old_ev:
                g.set_thick_line_width(c)
                c.set_source_rgb(self.__TraceColors[j][0],
                                 self.__TraceColors[j][1],
                                 self.__TraceColors[j][2])
                self.__line(g.ev(i, j, old_ev[0]), p)
            g.set_thin_line_width(c)
            c.set_source_rgb(self.__EndColor[0], self.__EndColor[1],
                             self.__EndColor[2])
            self.__cross(p, g.ev_size())

    def finalize(self):
        c, g = self.__ctx, self.__geo

        x  = g.trace_row_right(0, 0)[0]
        y1, yb = g.vline_ys()
        y2 = g.trace_row_right(self.__n - 1, 0)[1]
        g.set_thin_line_width(c)
        c.set_source_rgb(self.__FrameColor[0], self.__FrameColor[1],
                         self.__FrameColor[2])
        self.__line((x, y1), (x, y2))
        self.__text((g.ev_row_right(0, 0)[0], yb),
                    str(self.__n / g.addrs_per_page() + 1))


