/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <iostream>
#include <sstream>
#include <stdexcept>
#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

#include <wipal/tool/args.hh>
#include <wipal/tool/hexdump.hh>
#include <wipal/pcap/list.hh>
#include <wipal/filter/timetracker.hh>
#include <wipal/filter/non_noisy_prism.hh>

namespace
{

  std::ostream&
  operator << (std::ostream& os, struct timeval& tv)
  {
    return os << wpl::tool::microseconds (tv) << '\n';
  }

  void
  print_colheads(const wpl::opt::list& opts, const std::string& name)
  {
    using namespace wpl;

    const opt::list&		ui = opts["ui"].get<opt::list>();
    const bool&			indices = ui["indices"].get<bool>();
    const bool&			microseconds = ui["microseconds"].get<bool>();
    const args::hex_summary&	hex_summary = ui["hex_summary"]
						.get<args::hex_summary>();

    std::ostringstream	l1;
    std::ostringstream	l2;

    std::cout << name << '\n';

    if (indices)
      {
	l1 << "Fr. ID";
	l2 << "======";
      }

    if (microseconds)
      {
	if (indices)
	  {
	    l1 << '\t';
	    l2 << '\t';
	  }
	l1 << "Microseconds";
	l2 << "============";
      }

    switch (hex_summary)
      {
      case args::short80211:
	if (indices or microseconds)
	  {
	    l1 << '\t';
	    l2 << '\t';
	  }
	l1 << "IEEE 802.11 hex. dump";
	l2 << "=====================";
	break;

      default:
	break;
      }

    std::cout << l1.str() << '\n' << l2.str() << std::endl;
  }

  template <class Phy, class Frame>
  void
  list_frame(const Frame&			f,
	     const bool&			indices,
	     const bool&			microseconds,
	     const wpl::args::hex_summary&	content)
  {
    using namespace wpl;
    namespace hex = tool::hex;
    namespace fmt = tool::fmt;

    if (indices)
      std::cout << f.meta().id;

    if (microseconds)
      {
	if (indices)
	  std::cout << '\t';
	std::cout << f.microseconds();
      }

    switch (content)
      {
      case args::short80211:
	{
	  const Phy* const	phy = static_cast<const Phy*> (f.bytes());
	  const bool		swapped = f.swapped();
	  const size_t		caplen = f.meta().caplen;
	  const void* const	ieee = phy->decapsulate(caplen, swapped);
	  const size_t		ieee_len = caplen - phy->len(caplen, swapped);

	  if (indices or microseconds)
	    std::cout << '\t';
	  std::cout << hex::dump<hex::params::str>(ieee, ieee_len,
						   fmt::printable ()) << '\n';
	}
	break;

      case args::full:
	if (indices or microseconds)
	  std::cout << '\n';
	std::cout << hex::dump<hex::params::def>(f.bytes(), f.meta().caplen,
						 fmt::printable ());
	break;

      case args::nohex:
	if (indices or microseconds)
	  std::cout << '\n';
	break;

      default:
	throw std::logic_error ("Internal error: unreachable code. "
				"Please fill a bug report.");
	break;
      }
    std::cout << std::flush;
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      using namespace wpl;
      namespace def = args::def;

      typedef tool::end::endianness		endianness;

      opt::list opts ((opt::name ("internals") -= opt::list
		       (( opt::name ("endianness") -= def::endianness(),
			  opt::name ("ignore_noisy_prism") -=
						     def::ignore_noisy_prism(),
			  opt::name ("phy_tstamps") -= def::phy_tstamps())),
		       opt::name ("ui") -= opt::list
		       ((opt::name ("headers") -= def::ui_headers(),
			 opt::name ("indices") -= def::ui_indices(),
			 opt::name ("microseconds") -= def::ui_microseconds(),
			 opt::name ("hex_summary") -= def::ui_hex_summary()))));

      args::parse(argc, argv, opts, args::arbitrary, args::none);

      BOOST_FOREACH(pcap::list<> t,
		    opt::get<args::input>(opts["proc"]["input"]))
	{
	  if (t.empty())
	    continue;

	  const pkt::type lt = t.type();
	  args::check_options(opts, lt);
	  t.expect(lt);

	  if (opt::get<bool>(opts["ui"]["headers"]))
	    print_colheads(opts, boost::lexical_cast<std::string> (t));

	  const opt::list&	ui = opt::get<opt::list>(opts["ui"]);
	  const endianness&	end = opt::get<endianness>(opts["internals"]
							   ["endianness"]);
	  const bool		pcapts = not opt::get<bool>(opts["internals"]
							    ["phy_tstamps"]);

	  switch (lt)
	    {

#define WP_case(R, Unused, LT)						\
	    case pkt::BOOST_PP_TUPLE_ELEM(2, 0, LT):			\
	      {								\
		using boost::bind;					\
									\
		typedef BOOST_PP_TUPLE_ELEM(2, 1, LT)	phy_header;	\
		typedef pcap::list<>::const_iterator	iterator;	\
		typedef filter::timetracker<iterator, phy_header>	\
							timetracker;	\
									\
		timetracker trk (t.begin(), t.end(), end, pcapts);	\
									\
		if (opt::get<bool>(opts["internals"]["ignore_noisy_prism"])) \
		  {							\
		    typedef timetracker::const_iterator		it;	\
		    typedef filter::non_noisy_prism<it>		nnp;	\
		    typedef nnp::const_iterator::value_type	frame;	\
									\
		    nnp	n (trk.begin(), trk.end());			\
									\
		    n.for_each(bind(&list_frame<phy_header, frame>, _1,	\
				    opt::get<bool>(ui["indices"]),	\
				    opt::get<bool>(ui["microseconds"]), \
				    opt::get<args::hex_summary>		\
				      (ui["hex_summary"])));		\
		  }							\
		else							\
		  {							\
		    typedef						\
		      timetracker::const_iterator::value_type		\
		      frame;						\
									\
		    trk.for_each(bind(&list_frame<phy_header, frame>, _1, \
				      opt::get<bool>(ui["indices"]),	\
				      opt::get<bool>(ui["microseconds"]), \
				      opt::get<args::hex_summary>	\
				      (ui["hex_summary"])));		\
		  }							\
	      }								\
	    break;

	      BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

	    default:
	      args::unsupported_linktype(lt);
	      break;
	    }
	}
    }
  catch (std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
