/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#include <cmath>
#include <fstream>
#include <algorithm>

#include <boost/format.hpp>

#include <wipal/tool/si.hh>

#include "wifi_stats.hh"
#include "config.hh"

namespace stats
{

  /*--------.
  | Helpers |
  `--------*/

  namespace
  {

    std::ostream&
    reinit(std::ostream&	os,
	   std::streamsize	p,
	   std::streamsize	w,
	   std::ios::fmtflags	f)
    {
      using namespace std;

      return os << setprecision (p)
		<< setw(w)
		<< resetiosflags (~f)
		<< setiosflags (f);
    }

    struct dump_functor
    {
      dump_functor(std::ostream& os, bool newline = true):
	os_ (os),
	nl_ (newline)
      {
      }

      std::ostream&
      operator () (const boost::posix_time::time_duration& x) const
      {
	const std::streamsize p = os_.precision();

	os_ << std::setprecision (6)
	    << (double (x.total_microseconds()) / 1e6)
	    << std::setprecision (p);
	if (nl_)
	  os_ << '\n';
	return os_;
      }

    private:
      std::ostream&	os_;
      bool		nl_;
    };

    struct dump_iomanipulator
    {
      typedef boost::posix_time::time_duration	time_duration;

      dump_iomanipulator(const time_duration& x, bool newline = false):
	td_ (x),
	nl_ (newline)
      {
      }

      std::ostream&
      dump(std::ostream& os) const
      {
	return dump_functor (os, nl_) (td_);
      }

    private:
      const time_duration&	td_;
      bool			nl_;
    };

    std::ostream&
    operator << (std::ostream& os, const dump_iomanipulator& dm)
    {
      return dm.dump(os);
    }

  } // End of anonymous namespace.


    /*-------------.
    | distribution |
    `-------------*/

  std::ostream&
  distribution::dump(std::ostream& os) const
  {
    for (unsigned i = 0; i < dist_.size(); ++i)
      os << (min_ + gran_ / 2 + (i - 1) * gran_) << '\t'
	 << (float (dist_[i]) / float (total_)) << '\n';
    return os << std::flush;
  }

  /*---------------.
  | probe_requests |
  `---------------*/

  const float		probe_requests::unit	= 1e6;
  const float		probe_requests::min	= 0 * unit;
  const float		probe_requests::max	= 300 * unit;
  const float		probe_requests::gran	= 5 * unit;
  const unsigned	probe_requests::tic	= 4;

  std::ostream&
  probe_requests::dump_header(std::ostream&	os,
			      unsigned	successful_frames) const
  {
    return os << "* begin Probe Requests\n"
	      << "count: " << count << '\n'
	      << "ratio to successful frames: "
	      << std::setprecision (1)
	      << (100 * (float (count) / float (successful_frames)))
	      << "%\n" << std::endl;
  }

  std::ostream&
  probe_requests::dump_footer(std::ostream& os) const
  {
    return os << "* end Probe Requests" << std::endl;
  }

  std::ostream&
  probe_requests::dump_addr_stats(std::ostream&	os,
				  unsigned		i,
				  unsigned		distinct_transmitters)
    const
  {
    const addr_state_map&	m = *addrs;
    unsigned			len = 0;

    os << "  * begin " << wpl::wifi::mgt::header::addr_names[i] << '\n'
       << "  format: <id>   <addr> <count>   <mean IT>  <IT stddev>"
      "  <median IT>\n"
       << std::endl;

    for (addr_state_map::const_iterator it = m.begin();
	 it != m.end();
	 ++it)
      {
	const std::vector<ptime>& evs = it->second.pr_stats[i].evs();

	if (evs.empty())
	  continue;
	++len;
	os << std::setw (5) << len << ' ' << it->first << ' '
	   << std::setw (5) << evs.size();
	if (evs.size() < 2)
	  os << "            0            0            0\n";
	else
	  {
	    using boost::posix_time::time_duration;

	    const std::vector<time_duration> its_v = its(evs);

	    { // Do output the interarrival times sequence.
	      const std::string s =
		str(boost::format ("output/addr%d-%s.seq")	% (i + 1)
		    % it->first);
	      std::ofstream o (s.c_str());
	      std::for_each(its_v.begin(), its_v.end(), dump_functor (o));
	    }

	    { // Do output the interarrival times distribution.
	      const std::string s =
		str(boost::format ("output/addr%d-%s.dist")	% (i + 1)
		    % it->first);
	      std::ofstream o (s.c_str());
	      dump_it_dist(o, make_distribution(min, max, gran, its_v));
	    }

	    typedef dump_iomanipulator dumper;

	    os << ' ' << std::setw (12) << dumper (mean(its_v))
	       << ' ' << std::setw (12) << dumper (stddev(its_v))
	       << ' ' << std::setw (12) << dumper (median(its_v))
	       << '\n';
	  }
      }


    return os << "\n"
      "  length: " << len << '\n'
	      << "  ratio to distinct transmitters: "
	      << std::setprecision (1)
	      << (100 * (float (len) / float (distinct_transmitters)))
	      << "%\n"
	      << "  * end " << wpl::wifi::mgt::header::addr_names[i] << "\n\n"
	      << std::endl;
  }

  std::ostream&
  probe_requests::dump_pr_events(std::ostream& os,
				 unsigned i,
				 const ptime& ref) const
  {
    typedef addr_state_map::const_iterator	addrs_iterator;
    typedef std::vector<ptime>		vector;

    unsigned c = 0;

    for (addrs_iterator it = addrs->begin(); it != addrs->end(); ++it)
      {
	const vector&	evs = it->second.pr_stats[i].evs();

	if (not evs.empty())
	  ++c;

	for (vector::const_iterator ev = evs.begin(); ev != evs.end(); ++ev)
	  os << dump_iomanipulator (*ev - ref) << '\t' << c << '\n';
      }

    return os << std::flush;
  }

  std::ostream&
  probe_requests::dump_it_dist(std::ostream& os,
			       const distribution& dist)
  {
    os << "set xrange [" << (min - gran) << ':' << (max + gran) << "]\n"
      "set yrange [0:1]\n"
      "set xlabel 'Interarrival time between two probe requests (s)'\n"
      "set ylabel 'Frequency'\n"
      "set xtics (";
    for (float f = 0; f < max; f += tic * gran)
      os << '"' << unsigned (f / unit)<< "\" " << f << ", ";
    os << "\" >" << unsigned (max / unit) << "\" " << (max + gran / 2)
       << ")\n"
      "set ytics 0, 0.1\n"
      "plot '-' with boxes notitle" << std::endl
       << dist << 'e' << std::endl;
    return os;
  }

  std::ostream&
  probe_requests::dump(std::ostream&	os,
		       unsigned	successful_frames,
		       unsigned	distinct_transmitters,
		       const ptime&	first) const
  {
    const std::streamsize	default_p = os.precision();
    const std::streamsize	default_w = os.width();
    const std::ios::fmtflags	default_f = os.flags();

    os << std::right << std::fixed;

    dump_header(os, successful_frames);
    if (not count)
      {
	dump_footer(os);
	return reinit(os, default_p, default_w, default_f);
      }

    {
      std::ofstream pr_evs ("output/pr_events");
      dump_pr_events(pr_evs, wpl::wifi::mgt::header::sa, first);
    }

    {
      std::ofstream dist_f ("output/it_dist");
      dump_it_dist(dist_f, it_dist);
    }

    for (unsigned i = 0; i < 3; ++i)
      dump_addr_stats(os, i, distinct_transmitters);

    dump_footer(os);
    return reinit(os, default_p, default_w, default_f);
  }

  /*-------.
  | frames |
  `-------*/

  std::ostream&
  frames::dump_pr_answers(std::ostream& os,
			  const addr_state_map::const_iterator& it) const
  {
    typedef addr_state::presp_attr		presp_attr;
    typedef addr_state::preq_answer_set	preq_answer_set;
    typedef addr_state::pr_answers_t		pr_answers_t;

    const pr_answers_t&		 pr_answers = it->second.pr_answers;
    const pr_answers_t::const_iterator begin = pr_answers.begin();

    for (pr_answers_t::const_iterator i = begin; i != pr_answers.end(); ++i)
      {
	enum { sa =  wpl::wifi::mgt::header::sa };

	os << it->second.pr_stats[sa].evs()[i - begin] << ' '
	   << i->size() << ' ';
	for (preq_answer_set::const_iterator j = i->begin();
	     j != i->end();
	     ++j)
	  {
	    const bool last_was_acked =
	      not j->second.ack_arrival.is_not_a_date_time();

	    os << (j != i->begin() ? " " : "") << j->first
	       << '[' << j->second.retries << ( last_was_acked ? " " : "  ")
	       <<  j->second.acks << (last_was_acked ? "*" : "") << ']';
	    // 	      if (last_was_acked) // Debug.
	    // 		os << j->second.ack_arrival << ' ';
	  }
	os << '\n';
      }
    return os << std::flush;
  }

  void
  frames::dump_all_pr_answers(const std::string& prefix,
			      const std::string& suffix) const
  {
    for (addr_state_map::const_iterator it = addrs.begin();
	 it != addrs.end();
	 ++it)
      {
	if (it->second.pr_answers.empty())
	  continue;

	const std::string s = str(boost::format ("%s%s%s") % prefix
				  % it->first
				  % suffix);
	std::ofstream o (s.c_str());
	dump_pr_answers(o, it);
      }
  }

  std::ostream&
  frames::dump(std::ostream& os) const
  {
    using wpl::tool::B;
    using wpl::tool::MiB;

    const std::streamsize	default_p = os.precision();
    const std::streamsize	default_w = os.width();
    const std::ios::fmtflags	default_f = os.flags();

    dump_all_pr_answers("output/addr2-", ".pr_answers");

    os << std::right << std::fixed;

    os << "frame count: " << frame_count << std::endl;
    if (not frame_count)
      return reinit(os, default_p, default_w, default_f);

    os << "start time: " << first << '\n'
       << "stop time:  " << last << '\n'
       << "duration:   " << (last - first) << '\n'
       << "PHY-invalid frames: " << phy_error_count << '\n'
       << "PHY-invalid ratio:  "
       << std::setprecision (1)
       << (100 * (float (phy_error_count) / float (frame_count))) << "%\n"
       << "frame sizes total: " << MiB (B (total_len)) << " ("
       << B (total_len) << ")\n"
       << "captured size:     " << MiB (B (total_caplen)) << " ("
       << B (total_caplen) << ")\n"
       << "distinct mentioned addresses: " << addrs.size() << '\n';

    unsigned ap_len = 0;
    unsigned distinct_tmit = 0;

    for (addr_state_map::const_iterator it = addrs.begin();
	 it != addrs.end();
	 ++it)
      {
	if (it->second.mentioned_as_ap)
	  ++ap_len;
	if (it->second.heard_as_tmiter)
	  ++distinct_tmit;
      }

    os << "distinct transmitters: " << distinct_tmit << '\n'
       << std::endl
       << "* begin mentioned AP's\n"
       << "length: " << ap_len << "\n"
      "format:    <addr>\t<count>\n" << std::endl;
    for (addr_state_map::const_iterator it = addrs.begin();
	 it != addrs.end();
	 ++it)
      if (it->second.mentioned_as_ap)
	os << it->first << '\t' << it->second.mentioned_as_ap << '\n';
    os << "* end mentioned AP's\n\n" << std::endl;

    return reinit(pr.dump(os,
			  frame_count - phy_error_count,
			  distinct_tmit,
			  first),
		  default_p,
		  default_w,
		  default_f);
  }

  std::ostream&
  operator << (std::ostream& os, const frames& f)
  {
    return f.dump(os);
  }

} // End of namespace stats.

