/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#include <set>
#include <vector>
#include <map>
#include <utility>
#include <iostream>
#include <stdexcept>
#include <cstdlib>

#include <boost/format.hpp>
#include <boost/foreach.hpp>
#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include <wipal/tool/microseconds.hh>
#include <wipal/tool/microseconds_stamp.hh>
#include <wipal/tool/args.hh>
#include <wipal/wifi/addr_mapping.hh>
#include <wipal/pcap/list.hh>
#include <wipal/filter/uniquely_identifiable.hh>

namespace
{

  template <class UniqueID>
  class uniqueness_tester
  {

    struct collision
    {
      collision(const unsigned			rhs_id,
		const wpl::tool::microseconds&	interval):
	rhs_id (rhs_id),
	interval (interval)
      {
      }

      unsigned			rhs_id;
      wpl::tool::microseconds	interval;
    };

    typedef wpl::tool::microseconds_stamp<UniqueID>	arrival;
    typedef std::set<arrival>				set_of_arrivals;
    typedef std::vector<collision>			collisions_vector;
    typedef std::map<arrival, collisions_vector>	collisions_map;

    const wpl::opt::list&	ui_opts_;
    set_of_arrivals		arrivals_;
    collisions_map		collisions_;
    unsigned			arrival_count_;
    unsigned			collision_count_;

  public:

    uniqueness_tester(const wpl::opt::list& ui_opts):
      ui_opts_ (ui_opts),
      arrival_count_ (0),
      collision_count_ (0)
    {
    }

    void		operator () (const arrival& a)
    {
      typedef typename set_of_arrivals::iterator iterator;

      const std::pair<iterator, bool> r = arrivals_.insert(a);

      ++arrival_count_;
      if (not r.second)
	{
	  const wpl::tool::microseconds dt =
	    a.microseconds() - r.first->microseconds();

	  ++collision_count_;
	  collisions_[*r.first].push_back(collision (a.frame_id(), dt));
	}
    }

    void	print(std::ostream& columns, std::ostream& summary) const
    {
      typedef typename collisions_vector::const_iterator	cvec_iterator;
      typedef typename collisions_map::const_iterator		cmap_iterator;

      boost::format f ("%15s %15s %15s");

      if (ui_opts_["headers"].get<bool>())
	{
	  columns << (f % "frame 1" % "frame 2" % "interval (us)") << '\n';
	  columns << (f % "=======" % "=======" % "=============") << std::endl;
	}
      for (cmap_iterator i = collisions_.begin(); i != collisions_.end(); ++i)
	{
	  bool first = true;
	  for (cvec_iterator j = i->second.begin(); j != i->second.end(); ++j)
	    {
	      columns << (f % i->first % j->rhs_id % j->interval);
	      if (first)
		{
		  first = false;
		  columns << " (new)";
		}
	      columns << '\n';
	    }
	}
      if (ui_opts_["summary"].get<bool>())
	summary << arrival_count_ << " unique frames, "
		<< collision_count_ << " collisions." << std::endl;
    }

    unsigned	collision_count() const
    {
      return collision_count_;
    }

  };


  struct main_helper
  {
    template <class UniqueId>
    int
    operator () (wpl::opt::list& options) const
    {
      using namespace wpl;

      bool			success (true);

      BOOST_FOREACH(pcap::list<> traces,
		    opt::get<args::input>(options["proc"]["input"]))
	{
	  using filter::for_each_uniquely_identifiable;

	  uniqueness_tester<UniqueId>	tester (opt::get<opt::list>(options
								    ["ui"]));

	  if (opt::get<bool>(options["internals"]["ignore_noisy_prism"]))
	    traces.expect(pkt::PRISM_HEADER);

	  if (traces.empty())
	    continue;

	  const pkt::type	t = traces.get(0).type();
	  args::check_options(options, t);
	  traces.expect(t);

	  switch (t)
	    {

#define WP_case(R, Unused, LT)						\
	    case pkt::BOOST_PP_TUPLE_ELEM(2, 0, LT):			\
	      for_each_uniquely_identifiable<UniqueId,			\
					     BOOST_PP_TUPLE_ELEM(2, 1, LT)> \
	        (traces.begin(), traces.end(),				\
		 opt::get<opt::list>(options["internals"]), tester);	\
	      break;

	      BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)
#undef WP_case

	    default:
	      args::unsupported_linktype(t);
	    }

	  if (opt::get<bool>(options["ui"]["headers"]))
	    std::cout << traces << ":\n";
	  tester.print(std::cout, std::cerr);
	  if (tester.collision_count())
	    success = false;
	}

      return success ? EXIT_SUCCESS : EXIT_FAILURE;
    }
  }; // End of struct main_helper.

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      using namespace wpl;
      namespace def = args::def;

      opt::list	opts ((opt::name ("internals") -= opt::list
		       ((opt::name ("endianness") -= def::endianness(),
			 opt::name ("ignore_noisy_prism") -=
						     def::ignore_noisy_prism(),
			 opt::name ("mapping") -= def::mapping(),
			 opt::name ("phy_tstamps") -= def::phy_tstamps())),

		       opt::name ("misc") -= opt::list
		       ((opt::name ("mac_map") -= def::mac_map(),
			 opt::name ("attributes") -= def::attributes())),

		       opt::name ("ui") -= opt::list
		       ((opt::name ("headers") -= def::ui_headers(),
			 opt::name ("summary") -= def::ui_summary()))));

      args::parse(argc, argv, opts, args::arbitrary, args::none);
      return args::dispatch_unique_id(opts, main_helper ());
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
